/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.aed;

import com.alicecallsbob.fcsdk.android.ListenerTarget;

/**
 * A representation of a single Topic within AED. This object allows the
 * publishing and accessing of data within the topic.
 */
public interface Topic extends ListenerTarget<TopicListener>
{
    /**
     * Returns the name of the topic (which also uniquely identifies the topic).
     * 
     * @return the name of the topic.
     */
    String getName();

    /**
     * Whether the topic is connected or not.
     * 
     * @return <code>true</code> if the topic is connected, <code>false</code>
     *         otherwise.
     */
    boolean isConnected();

    /**
     * Disconnects from a topic. Does not delete the topic or it's data.
     */
    void disconnect();

    /**
     * Disconnects from a topic and optionally deletes the topic at the same
     * time.
     * 
     * @param deleteTopic Whether or not to delete the topic and it's data, as
     *            well as disconnect from it.
     * 
     */
    void disconnect(boolean deleteTopic);

    /**
     * Submits data to a topic.
     * 
     * @param key a unique identifier which is associated with a value.
     * 
     * @param value the single value that is associated to the specified key.
     * 
     */
    void submitData(String key, String value);

    /**
     * Deletes data from a topic.
     * 
     * @param key a unique identifier which is associated with a value.
     * 
     */
    void deleteData(String key);

    /**
     * Sends a custom message that will be received by any subscribers to the
     * specified topic.
     * 
     * @param aedMessage message to send
     * 
     */
    void sendAedMessage(String aedMessage);
}
