/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.aed.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.json.JSONObject;

import android.util.Log;

import com.alicecallsbob.fcsdk.android.aed.AED;
import com.alicecallsbob.fcsdk.android.aed.Topic;
import com.alicecallsbob.fcsdk.android.aed.TopicListener;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicCreatedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicDataHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicDeletedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicExpiredHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicMessageHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicMessageNotSentHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicMessageSentHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicNotCreatedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicNotDeletedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicNotPublishedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicNotSubscribedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicNotUnpublishedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicNotUnsubscribedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicPublishedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicSubscribedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicUnpublishedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicUnsubscribedHandler;
import com.alicecallsbob.fcsdk.android.aed.impl.handler.TopicUpdatedHandler;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageHandler;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageManager;

public class AEDManager implements AED
{
    private final static String  TAG = AEDManager.class.getSimpleName();
    private final ServerMessageManager serverMessageManager;
    private final List<TopicImpl> topics = new ArrayList<TopicImpl>();

    public AEDManager(ServerMessageManager serverMessageManager)
    {
        super();
        this.serverMessageManager = serverMessageManager;
        setupHandlers();
    }

    private void setupHandlers()
    {
        final TopicCreatedHandler topicCreatedHandler = new TopicCreatedHandler(this);
        final TopicDeletedHandler topicDeletedHandler = new TopicDeletedHandler(this);
        final TopicDataHandler topicDataHandler = new TopicDataHandler(this);
        final TopicPublishedHandler topicPublishedHandler = new TopicPublishedHandler(this);
        final TopicUnpublishedHandler topicUnpublishedHandler = new TopicUnpublishedHandler(this);
        final TopicSubscribedHandler topicSubscribedHandler = new TopicSubscribedHandler(this);
        final TopicUnsubscribedHandler topicUnsubscribedHandler = new TopicUnsubscribedHandler(this);
        final TopicMessageSentHandler topicMessageSentHandler = new TopicMessageSentHandler(this);
        final TopicNotCreatedHandler topicNotCreatedHandler = new TopicNotCreatedHandler(this);
        final TopicNotDeletedHandler topicNotDeletedHandler = new TopicNotDeletedHandler(this);
        final TopicNotPublishedHandler topicNotPublishedHandler = new TopicNotPublishedHandler(this);
        final TopicNotUnpublishedHandler topicNotUnpublishedHandler = new TopicNotUnpublishedHandler(this);
        final TopicNotSubscribedHandler topicNotSubscribedHandler = new TopicNotSubscribedHandler(this);
        final TopicNotUnsubscribedHandler topicNotUnsubscribedHandler = new TopicNotUnsubscribedHandler(this);
        final TopicMessageNotSentHandler topicMessageNotSentHandler = new TopicMessageNotSentHandler(this);
        final TopicExpiredHandler topicExpiredHandler = new TopicExpiredHandler(this);
        final TopicUpdatedHandler topicUpdatedHandler = new TopicUpdatedHandler(this);
        final TopicMessageHandler topicMessageHandler = new TopicMessageHandler(this);

        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_CREATED, topicCreatedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_DELETED, topicDeletedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_DATA, topicDataHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_PUBLISHED, topicPublishedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_UNPUBLISHED, topicUnpublishedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_SUBSCRIBED, topicSubscribedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_UNSUBSCRIBED, topicUnsubscribedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_MESSAGE_SENT, topicMessageSentHandler);
        // We only get a topic conflict when the topic already exists, so just
        // subscribe
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_CONFLICT, topicCreatedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_CREATE_ERROR, topicNotCreatedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_DELETE_ERROR, topicNotDeletedHandler);
        // The only time we retrieve data is on connect, so trigger a connect
        // failure.
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_DATA_ERROR, topicNotCreatedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_PUBLISH_ERROR, topicNotPublishedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_UNPUBLISH_ERROR, topicNotUnpublishedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_SUBSCRIBE_ERROR, topicNotSubscribedHandler);
        // This means we are already subscribed, and the user has attempted to
        // connect to a topic they are already connected to. Therefore we will
        // treat this the same as a successful subscribe.
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_SUBSCRIBE_CONFLICT, topicSubscribedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_UNSUBSCRIBE_ERROR, topicNotUnsubscribedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_SEND_MESSAGE_ERROR, topicMessageNotSentHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_EXPIRED, topicExpiredHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_UPDATE, topicUpdatedHandler);
        addHandler(AEDServerMessageBuilder.MESSAGE_TYPE_TOPIC_MESSAGE, topicMessageHandler);
    }

    private void addHandler(String messageType, ServerMessageHandler handler)
    {
        serverMessageManager.addHandler(messageType, handler);
    }

    private void sendServerMessage(JSONObject json)
    {
        serverMessageManager.sendServerMessage(json);
    }

    private TopicImpl findTopic(String topicName)
    {
        for (TopicImpl topic : topics)
        {
            if (topic.getName().equals(topicName))
            {
                return topic;
            }
        }
        return null;
    }

    @Override
    public Topic createTopic(String topicName, TopicListener listener)
    {
        return createTopic(topicName, 0, listener);
    }

    @Override
    public Topic createTopic(String topicName, int expiryTime, TopicListener listener)
    {
        TopicImpl topic = findTopic(topicName);
        if (topic == null)
        {
            topic = new TopicImpl(serverMessageManager, listener, topicName);
            topics.add(topic);
        }
        else
        {
            topic.addListener(listener);
        }

        if (topic.isConnected() == false)
        {
            connectToTopic(topicName, expiryTime);
        }
        return topic;
    }

    private void connectToTopic(String topicName, int expiryTime)
    {
        final Integer expiryObject = expiryTime == 0 ? null : expiryTime;
        final JSONObject json = AEDServerMessageBuilder.createCreateTopicMessage(topicName, expiryObject);
        sendServerMessage(json);
    }

    public void onTopicCreated(String name)
    {
        final JSONObject json = AEDServerMessageBuilder.createSubscribeMessage(name);
        sendServerMessage(json);
    }

    public void onTopicDeleted(String name, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            if (topic.isConnected())
            {
                topic.setConnected(false);
                for (TopicListener listener : topic.listeners())
                {
                    listener.onTopicDeleted(topic, message);
                    listener.onTopicDeletedRemotely(topic);
                }
            }
            Log.d(TAG, "onTopicDeleted removing topic: " + name);
            topics.remove(topic);
        }
    }

    public void onTopicData(String name, Map<String, Object> data)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null && topic.isConnected())
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicConnected(topic, data);
            }
        }
    }

    public void onTopicPublished(String name, String key, String value, int version)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null && topic.isConnected())
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicSubmitted(topic, key, value, version);
            }
        }
    }

    public void onTopicUnpublished(String name, String key, int version)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null && topic.isConnected())
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onDataDeleted(topic, key, version);
            }
        }
    }

    public void onTopicSubscribed(String name)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            topic.setConnected(true);
            JSONObject json = AEDServerMessageBuilder.createGetAllDataMessage(name);
            sendServerMessage(json);
        }
    }

    public void onTopicUnsubscribed(String name)
    {
        // No callback fired on UNSUBSCRIBE, but we need to set state to
        // disconnected
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            topic.setConnected(false);
        }
    }

    public void onTopicMessageSent(String name, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicSent(topic, message);
            }
        }
    }

    public void onTopicNotCreated(String name, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicNotConnected(topic, message);
            }
            Log.d(TAG, "onTopicNotCreated removing topic: " + name);
            topics.remove(topic);
        }
    }

    public void onTopicNotDeleted(String name, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicNotDeleted(topic, message);
            }
        }
    }

    public void onTopicNotPublished(String name, String key, String value, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicNotSubmitted(topic, key, value, message);
            }
        }
    }

    public void onTopicNotUnpublished(String name, String key, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onDataNotDeleted(topic, key, message);
            }
        }
    }

    public void onTopicNotSubscribed(String name, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            topic.setConnected(false);
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicNotConnected(topic, message);
            }
        }
    }

    public void onTopicNotUnsubscribed(String name)
    {
        // Happens if you try to unsubscribe from a topic that you aren't
        // actually subscribed to. We should be disconnected now though.
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            topic.setConnected(false);
        }
    }

    public void onTopicMessageNotSent(String name, String message, String originalMessage)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            topic.setConnected(false);
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicNotSent(topic, originalMessage, message);
            }
        }
    }

    public void onTopicExpired(String name)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null)
        {
            topic.setConnected(false);
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicDeletedRemotely(topic);
            }
            Log.d(TAG, "onTopicExpired removing topic: " + name);
            topics.remove(topic);
        }
    }

    public void onTopicUpdated(String name, String key, String value, int version, boolean deleted)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null && topic.isConnected())
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onTopicUpdated(topic, key, value, version, deleted);
            }
        }
    }

    public void onTopicMessage(String name, String message)
    {
        final TopicImpl topic = findTopic(name);
        if (topic != null && topic.isConnected())
        {
            for (TopicListener listener : topic.listeners())
            {
                listener.onMessageReceived(topic, message);
            }
        }
    }
}
