/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.aed.impl;

import org.json.JSONObject;

import com.alicecallsbob.fcsdk.android.impl.ServerMessageBuilder;

public abstract class AEDServerMessageBuilder extends ServerMessageBuilder
{
    // Sent Message Types
    public static final String MESSAGE_TYPE_CREATE_TOPIC = "CREATE_TOPIC";
    public static final String MESSAGE_TYPE_DELETE_TOPIC = "DELETE_TOPIC";
    public static final String MESSAGE_TYPE_GET_ALL_TOPIC_DATA = "GET_ALL_TOPIC_DATA";
    public static final String MESSAGE_TYPE_GET_TOPIC_DATA = "GET_TOPIC_DATA";
    public static final String MESSAGE_TYPE_PUBLISH = "PUBLISH";
    public static final String MESSAGE_TYPE_UNPUBLISH = "UNPUBLISH";
    public static final String MESSAGE_TYPE_SUBSCRIBE = "SUBSCRIBE";
    public static final String MESSAGE_TYPE_UNSUBSCRIBE = "UNSUBSCRIBE";
    public static final String MESSAGE_TYPE_SEND_MESSAGE = "SEND_MESSAGE";

    // Received Success Message Types
    public static final String MESSAGE_TYPE_SUBSCRIBED = "SUBSCRIBED";
    public static final String MESSAGE_TYPE_UNSUBSCRIBED = "UNSUBSCRIBED";
    public static final String MESSAGE_TYPE_PUBLISHED = "PUBLISHED";
    public static final String MESSAGE_TYPE_UNPUBLISHED = "UNPUBLISHED";
    public static final String MESSAGE_TYPE_MESSAGE_SENT = "MESSAGE_SENT";
    public static final String MESSAGE_TYPE_TOPIC_CREATED = "TOPIC_CREATED";
    public static final String MESSAGE_TYPE_TOPIC_DELETED = "TOPIC_DELETED";
    public static final String MESSAGE_TYPE_TOPIC_DATA = "TOPIC_DATA";
    public static final String MESSAGE_TYPE_TOPIC_EXPIRED = "TOPIC_EXPIRED";
    public static final String MESSAGE_TYPE_TOPIC_UPDATE = "TOPIC_UPDATE";
    public static final String MESSAGE_TYPE_TOPIC_MESSAGE = "TOPIC_MESSAGE";

    // Received Error Message Types
    public static final String MESSAGE_TYPE_TOPIC_CONFLICT = "TOPIC_CONFLICT";
    public static final String MESSAGE_TYPE_TOPIC_CREATE_ERROR = "TOPIC_CREATE_ERROR";
    public static final String MESSAGE_TYPE_TOPIC_DELETE_ERROR = "TOPIC_DELETE_ERROR";
    public static final String MESSAGE_TYPE_TOPIC_DATA_ERROR = "TOPIC_DATA_ERROR";
    public static final String MESSAGE_TYPE_PUBLISH_ERROR = "PUBLISH_ERROR";
    public static final String MESSAGE_TYPE_UNPUBLISH_ERROR = "UNPUBLISH_ERROR";
    public static final String MESSAGE_TYPE_SUBSCRIBE_ERROR = "SUBSCRIBE_ERROR";
    public static final String MESSAGE_TYPE_SUBSCRIBE_CONFLICT = "SUBSCRIBE_CONFLICT";
    public static final String MESSAGE_TYPE_UNSUBSCRIBE_ERROR = "UNSUBSCRIBE_ERROR";
    public static final String MESSAGE_TYPE_SEND_MESSAGE_ERROR = "SEND_MESSAGE_ERROR";

    public static final String MESSAGE_FIELD_DATA = "data";
    public static final String MESSAGE_FIELD_MESSAGE = "message";
    public static final String MESSAGE_FIELD_ORIGINAL_MESSAGE = "original_message";
    public static final String MESSAGE_FIELD_VERSION = "version";
    public static final String MESSAGE_FIELD_TOPIC = "topic";
    public static final String MESSAGE_FIELD_KEY = "key";
    public static final String MESSAGE_FIELD_VALUE = "value";
    public static final String MESSAGE_FIELD_TIMEOUT = "timeout";
    public static final String MESSAGE_FIELD_DELETED = "deleted";

    public static JSONObject createCreateTopicMessage(String topicName, Integer expiry)
    {
        return createMessageWithType(MESSAGE_TYPE_CREATE_TOPIC, MESSAGE_FIELD_TOPIC, topicName, MESSAGE_FIELD_TIMEOUT,
                expiry);
    }

    public static JSONObject createDeleteTopicMessage(String topicName)
    {
        return createMessageWithType(MESSAGE_TYPE_DELETE_TOPIC, MESSAGE_FIELD_TOPIC, topicName);
    }

    public static JSONObject createGetAllDataMessage(String topicName)
    {
        return createMessageWithType(MESSAGE_TYPE_GET_ALL_TOPIC_DATA, MESSAGE_FIELD_TOPIC, topicName);
    }

    public static JSONObject createPublishDataMessage(String topicName, String key, String value)
    {
        return createMessageWithType(MESSAGE_TYPE_PUBLISH, MESSAGE_FIELD_TOPIC, topicName, MESSAGE_FIELD_KEY, key,
                MESSAGE_FIELD_VALUE, value);
    }

    public static JSONObject createDeleteDataMessage(String topicName, String key)
    {
        return createMessageWithType(MESSAGE_TYPE_UNPUBLISH, MESSAGE_FIELD_TOPIC, topicName, MESSAGE_FIELD_KEY, key);
    }

    public static JSONObject createSubscribeMessage(String topicName)
    {
        return createMessageWithType(MESSAGE_TYPE_SUBSCRIBE, MESSAGE_FIELD_TOPIC, topicName);
    }

    public static JSONObject createUnsubscribeMessage(String topicName)
    {
        return createMessageWithType(MESSAGE_TYPE_UNSUBSCRIBE, MESSAGE_FIELD_TOPIC, topicName);
    }

    public static JSONObject createAedMessage(String topicName, String aedMessage)
    {
        return createMessageWithType(MESSAGE_TYPE_SEND_MESSAGE, MESSAGE_FIELD_TOPIC, topicName, MESSAGE_FIELD_MESSAGE,
                aedMessage);
    }
}
