/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.aed.impl.handler;

import java.util.Map;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ArrayList;

import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONArray;

import com.alicecallsbob.fcsdk.android.aed.impl.AEDManager;
import com.alicecallsbob.fcsdk.android.aed.impl.AEDServerMessageBuilder;

public class TopicDataHandler extends AEDServerMessageHandler
{
    public TopicDataHandler(AEDManager aedManager)
    {
        super(aedManager);
    }

    @Override
    public void handleServerMessage(JSONObject serverMessage) throws JSONException
    {
        final String name = serverMessage.getString(AEDServerMessageBuilder.MESSAGE_FIELD_TOPIC);
        final JSONObject json = serverMessage.optJSONObject(AEDServerMessageBuilder.MESSAGE_FIELD_DATA);
        final Map<String, Object> data = mapForJson(json);
        getAEDManager().onTopicData(name, data);
    }

    /**
     * Converts a JSON object to a {@link Map}. Any JSONObjects within are also
     * converted to Maps, and any JSONArrays are converted to ArrayLists.
     *
     * @param json The JSON object or null
     * @return A map that represents the JSON (or null if that was the input).
     * @throws JSONException if the JSON can't be converted.
     */
    private Map<String, Object> mapForJson(JSONObject json) throws JSONException
    {
        if (json != null)
        {
            return convertObject(json);
        }
        return null;
    }

    /**
     * Convert the provided JSONObject to a {@link Map}. Convert any
     * JSONObject or JSONArray values to Map or {@link ArrayList}
     * as appropriate.
     * 
     * @param json The JSONObject to convert.
     * @return A new Map as specified in the description.
     * @throws JSONException if the JSONObject cannot be converted.
     */
    private Map<String, Object> convertObject(JSONObject json) throws JSONException
    {
        Map<String, Object> jsonContents = new LinkedHashMap<String, Object>();
        Iterator<?> keys = json.keys();
        while (keys.hasNext())
        {
            String key = (String)keys.next();
            Object value = convertIfNeeded(json.get(key));
            jsonContents.put(key, value);
        }
        return jsonContents;
    }

    /**
     * Convert any JSONObject or JSONArray to {link Map} or
     * {link ArrayList} as appropriate. If the object is not one of
     * the above types, return it as it is.
     * 
     * @param object The object to convert if needed.
     * @return The object converted if needed.
     * @throws JSONException if the JSONObject or JSONArray cannot be converted.
     */
    private Object convertIfNeeded(Object object) throws JSONException
    {
        if (object instanceof JSONObject)
        {
            object = convertObject((JSONObject) object);
        }
        else if (object instanceof JSONArray)
        {
            object = convertObject((JSONArray) object);
        }

        return object;
    }

    /**
     * Convert the provided JSONArray to an {@link ArrayList}. Convert any
     * JSONObject or JSONArray values to {@link Map} or ArrayList
     * as appropriate.
     * 
     * @param value The JSONArray to convert.
     * @return A new ArrayList as specified in the description.
     * @throws JSONException if the JSONArray cannot be converted.
     */
    private List<Object> convertObject(JSONArray json) throws JSONException
    {
        ArrayList<Object> jsonContents = new ArrayList<Object>(json.length());
        
        for (int index = 0; index < json.length(); index++)
        {
            Object value = convertIfNeeded(json.get(index));
            jsonContents.add(value);
        }
        
        return jsonContents;
    }

}
