/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.impl;

import org.json.JSONException;
import org.json.JSONObject;

public abstract class ServerMessageBuilder
{
    public static final String MESSAGE_TYPE_SYSTEM_FAILURE = "SYSTEM_FAILURE";
    public static final String MESSAGE_TYPE_MULTIPART = "MULTIPART";

    public static final String MESSAGE_FIELD_TYPE = "type";
    public static final String MESSAGE_FIELD_MUTIPART_ID = "multipartID";
    public static final String MESSAGE_FIELD_FINAL_MUTIPART_ID = "finalMultipartID";
    public static final String MESSAGE_FIELD_MULTIPART_CONTENT = "content";

    protected static JSONObject createMessageWithType(String type)
    {
        return createMessageWithType(type, new Object[] {});
    }

    /**
     * Creates a JSON representation of the given type and key/value pairs.
     * 
     * @param type the message type (cannot be <code>null</code> or empty).
     * @param args the key/value pairs - must be an even number of arguments in
     *            the form key1, val1, key2, val2. Duplicate keys are not
     *            allowed and keys cannot be <code>null</code>.
     * @throws IllegalArgumentException If the arguments are invalid (see above)
     *             or a {@link JSONException} is thrown.
     * @return the JSON object.
     */
    protected static JSONObject createMessageWithType(String type, Object... args)
    {
        if (type == null || type.length() == 0)
        {
            // Don't allow null or empty type...
            throw new IllegalArgumentException("Type must not be empty or null.");
        }

        // There must be an even number of params...
        if (args.length % 2 != 0)
        {
            throw new IllegalArgumentException(String.format("Must be an even number of args - found %d.", args.length));
        }

        try
        {
            final JSONObject json = new JSONObject();
            json.put(ServerMessageBuilder.MESSAGE_FIELD_TYPE, type);
            for (int i = 0; i < args.length; i += 2)
            {
                final Object key = args[i];
                final Object value = args[i + 1];
                if (key != null)
                {
                    final String keyString = key.toString();
                    if (json.has(keyString))
                    {
                        throw new IllegalArgumentException(String.format("Multiple values found for key - %s.",
                                keyString));
                    }
                    else
                    {
                        json.put(keyString, value);
                    }
                }
                else
                {
                    throw new IllegalArgumentException("Keys cannot be null.");
                }
            }
            return json;
        }
        catch (JSONException ex)
        {
            throw new IllegalArgumentException(ex);
        }
    }

    public static JSONObject createMultipartMessage(int index, int total, String content)
    {
        return createMessageWithType(MESSAGE_TYPE_MULTIPART, MESSAGE_FIELD_MUTIPART_ID, index,
                MESSAGE_FIELD_FINAL_MUTIPART_ID, total, MESSAGE_FIELD_MULTIPART_CONTENT, content);
    }
}
