/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.impl;

import java.io.IOException;
import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.CookieStore;
import java.net.HttpCookie;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.locks.ReentrantLock;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.TrustManager;

import org.apache.http.message.BasicNameValuePair;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.os.AsyncTask;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;

import com.codebutler.android_websockets.WebSocketClient;


public class ServerMessageTransport implements WebSocketClient.Listener
{
	private static final String	TAG = "ServerMessageTransport";

    private static final int MULTIPART_MESSAGE_MINIMUM_SIZE = 700;
    private static final int MULTIPART_MESSAGE_CHUNK_SIZE = 500;

    /** Periods to wait between reconnection attempts */
    private static final int[] RECONNECT_PERIODS = {500, 1000, 2000, 4000, 4000, 4000, 4000};

	private static final int CLOSE_CANNOT_CONNECT = 0;

	private static final int CLOSE_CONNECTION_LOST = -2;
	
	private static final int CLOSE_DISCONNECTED = -4;
	
	private static final String COOKIE_HEADER = "Cookie";
    
    /** The transport listener. */
    private ServerMessageTransportListener listener;
    /** The websocket. */
    //private WebSocketConnection webSocket;
    private WebSocketClient webSocket;
    /** The URL to the websocket. */
    private String url;
    /** Whether the network is reachable. */
    private boolean reachable = false;
    /** We need to inform the listener when the websocket opens, but only the 1st time. */
    private boolean initialWebsocketOpened = false;
    /** Flag to indicate the web socket is being deliberately closed */
    private boolean closing = false;
    /** The last multipart id received. */
    private int lastMultipartId;
    /** The multipart message being built up. */
    private final StringBuilder multipartMessageBuilder = new StringBuilder();
    /** List of messages that are queued for sending once the WebSocket is connected */
    private final Queue<JSONObject> queuedMessages = new ArrayDeque<JSONObject>();
    /** Lock to ensure that no more than one thread concurrently attempts a (re)connection */
    private final ReentrantLock connectAttemptLock = new ReentrantLock();
    /** How many reconnection attempts have already been made */
    private int reconnectAttempt = 0;
    /** A websocket has been reconnected and we are waiting for a message to confirm its up (usually a PING) */
    private boolean waitingForMessageAfterReconnect = false;
    /** Send cookies with a websocket from the default cookie manager when making a connection */
    private boolean useCookies = false;

    private InetAddress[] lookedUpAddresses;
    private String host;
    
    private TrustManager trustManager;

	private HostnameVerifier hostnameVerifier;

    protected WebSocketClient createWebSocketInstance()
    {
        final URI uri = URI.create(url);
    	return new WebSocketClient(uri, this, getExtraHeaders(uri));
    }

    
    private BasicNameValuePair getCookieHeaders(final URI uri)
    {
        //Extract cookies and add as headers
        final CookieManager cookieManager = (CookieManager) CookieManager.getDefault();
        
        if (cookieManager != null && useCookies)
        {
            final CookieStore store = cookieManager.getCookieStore();
            final List<HttpCookie> cookies = store.get(uri);
            
            if (cookies != null && !cookies.isEmpty())
            {
                final StringBuilder builder = new StringBuilder();
                int idx = 0;
                for (final HttpCookie cookie:cookies)
                {
                    if (idx++ > 0)
                    {
                        builder.append(";");
                    }

                    builder.append(cookie.toString());
                    
                }
                return new BasicNameValuePair(COOKIE_HEADER, builder.toString());
            }
            else
            {
                Log.d(TAG, "Unable to retrieve any cookies");
            }
        }
        else
        {
            Log.d(TAG, "No default cookie handler has been defined");
        }
        return null;
    }
    
    private List<BasicNameValuePair> getExtraHeaders(final URI uri)
    {
        
        final ArrayList<BasicNameValuePair> headerList = new ArrayList<BasicNameValuePair>();

        final BasicNameValuePair agentHeader = new BasicNameValuePair("User-Agent", getUserAgentHeaderValue());
        headerList.add(agentHeader);

        final BasicNameValuePair cookies = getCookieHeaders(uri);
        if (cookies != null)
        {
            headerList.add(cookies);
        }
        return headerList;
    }
    
    private String getUserAgentHeaderValue()
    {
    	return new StringBuilder(android.os.Build.MANUFACTURER)
    		.append(" ")
    		.append(android.os.Build.MODEL)
    		.append(" with Android ")
    		.append(android.os.Build.VERSION.RELEASE)
    		.toString();
    }
    
    public ServerMessageTransportListener getListener()
    {
        return listener;
    }

    public void setListener(ServerMessageTransportListener listener)
    {
        this.listener = listener;
    }

    public void open(String url)
    {
    	Log.d(TAG, "open");
        this.url = url;

    	System.out.println("ServerMessageTransport.open " + url);

        if (reachable)
        {
            connectAttemptLock.lock();
            webSocket = createWebSocketInstance();
           
            WebSocketClient.setTrustManager(trustManager);            
            WebSocketClient.setHostnameVerifier(hostnameVerifier);

            try
            {
                webSocket.connect();
            }
            finally
            {
                connectAttemptLock.unlock();
            }
        }
        else
        {
        	Log.i(TAG, "network is not reachable");
        	onDisconnect(CLOSE_CANNOT_CONNECT, "Network unreachable");
        }
    }

    public void setNetworkReachable(boolean reachable)
    {        
        if (this.reachable == reachable)
        {
            return;
        }
        
        Log.i(TAG, "Reachability changed to " + reachable);

        this.reachable = reachable;
        if (reachable && url != null)
        {
            reconnectAttempt = 0;
            open(url);
        }
        else if (!reachable)
        {
        	if (webSocket != null)
        	{
        		webSocket.disconnect();
        		webSocket = null;
        	}
        }
    }
    
    public void setUseCookies(final boolean shouldUseCookies)
    {
        if(useCookies == shouldUseCookies)
        {
            return;
        }
        
        useCookies = shouldUseCookies;
        Log.i(TAG, "Use cookies changed to " + useCookies);
    }

    /**
     * Attempts to send a JSON message to the server, adding it to the message
     * queue if this is not immediately possible.
     *
     * @param serverMessage
     *            the message to send
     * @return <code>true</code> if the message was successfully sent,
     *         <code>false</code> if it was added to the message queue
     */
    public boolean sendServerMessage(JSONObject serverMessage)
    {
        Log.d(TAG, "sendServerMessage");
        Log.d(TAG, "Send: " + serverMessage.toString());
        if (webSocket != null)
        {
            if (webSocket.isConnected())
            {
                final String jsonString = serverMessage.toString();
                if (jsonString.length() > MULTIPART_MESSAGE_MINIMUM_SIZE)
                {
                    sendMultipartServerMessage(jsonString);
                }
                else
                {
                    Log.d(TAG, "sendTextMessage");
                    webSocket.send(jsonString);
                }
                return true;
            }
            else
            {
                /* NW
                Log.w(TAG, "The web socket is not connected; message queued for sending on reconnection");
                synchronized (queuedMessages)
                {
                    queuedMessages.add(serverMessage);
                }*/
                onDisconnect(CLOSE_CONNECTION_LOST, "Connection lost, reason unknown");
                return false;
            }
        }
        else
        {
            /* NW
            Log.w(TAG, "The web socket is null; message queued for sending on connection");
            synchronized (queuedMessages)
            {
                queuedMessages.add(serverMessage);
            }*/
            if (!reachable)
            {
                Log.d(TAG, "Cannot send message as network reachability set to false: " + serverMessage.toString());
            }
            else
            {
                Log.d(TAG, "Cannot send message as network connection timed out: " + serverMessage.toString());
            }

            return false;
        }
    }

    public void close()
    {
        initialWebsocketOpened = false;
        closing = true;
        if (webSocket != null)
        {
            webSocket.disconnect();
        }
        url = null;
        webSocket = null;
    }

    private void sendMultipartServerMessage(String serverMessageString)
    {
    	Log.d(TAG, "sendMultipartServerMessage");
        final ArrayList<String> chunks = new ArrayList<String>();
        while (serverMessageString.length() > MULTIPART_MESSAGE_CHUNK_SIZE)
        {
            chunks.add(serverMessageString.substring(0, MULTIPART_MESSAGE_CHUNK_SIZE));
            serverMessageString = serverMessageString.substring(MULTIPART_MESSAGE_CHUNK_SIZE,
                    serverMessageString.length());
        }
        chunks.add(serverMessageString);

        final int total = chunks.size();
        int i = 1; // Chunks are numbered 1..n
        for (String content : chunks)
        {
            final JSONObject multipartMessage = ServerMessageBuilder.createMultipartMessage(i++, total, content);
            sendServerMessage(multipartMessage);
        }
    }

    @Override
    public void onConnect()
    {
        Log.d(TAG, "web socket open " + initialWebsocketOpened);

        if (initialWebsocketOpened == false)
        {
            reconnectAttempt = 0;
            initialWebsocketOpened = true;
            if (listener != null)
            {
                listener.onTransportOpened();
            }
        }
        else
        {
            //Don't reset connection attempts until a message has been received (so we are sure
            //the web socket is working correctly.)
            waitingForMessageAfterReconnect = true;
        }

        // NW : Disable sending of queued messages as part of investigation of
        //      inability to make several (incoming and outgoing) calls .
        /*
        // Send queued messages
        synchronized (queuedMessages)
        {
            for (JSONObject message = queuedMessages.poll(); message != null; message = queuedMessages.poll())
            {
                if (!sendServerMessage(message))
                {
                    // Stop if the message can't be sent, as it will end up
                    // re-added to the queue and the loop will go on forever
                    break;
                }
            }
        }*/
    }

    @Override
    public void onMessage(String payload)
    {
        Log.d(TAG, "Received: " + payload);
        final JSONArray array = getMessages(payload);

        if (waitingForMessageAfterReconnect)
        {
            //We have reconnected are are waiting for a message to confirm that the websocket connection
            //is working.  This is usually the PING message (sent by the server when a connection is established).
            reconnectAttempt = 0;
            waitingForMessageAfterReconnect = false;

            if (listener != null)
            {
                listener.onTransportReestablished();
            }
        }
        if (array != null)
        {
            for (int i = 0; i < array.length(); i++)
            {
                try
                {
                    final JSONObject message = array.getJSONObject(i);
                    final String messageType = message.getString(ServerMessageBuilder.MESSAGE_FIELD_TYPE);
                    if (messageType.equals(ServerMessageBuilder.MESSAGE_TYPE_MULTIPART))
                    {
                        onMultipartMessage(message);
                    }
                    else
                    {
                        if (listener != null)
                        {
                            listener.onServerMessage(message);
                        }
                    }
                }
                catch (JSONException ex)
                {
                    // TODO: LOG error with json
                }
            }
        }
    }

    private JSONArray getMessages(String payload)
    {
        try
        {
            return new JSONArray(payload);
        }
        catch (JSONException ex)
        {
            final JSONArray array = new JSONArray();
            try
            {
                final JSONObject json = new JSONObject(payload);
                array.put(json);
                return array;
            }
            catch (JSONException innerEx)
            {
                // TODO: LOG error with json
                return null;
            }
        }
    }

    private void onMultipartMessage(JSONObject multipart) throws JSONException
    {
        final int currentChunk = multipart.getInt(ServerMessageBuilder.MESSAGE_FIELD_MUTIPART_ID);
        final int totalChunks = multipart.getInt(ServerMessageBuilder.MESSAGE_FIELD_FINAL_MUTIPART_ID);
        final String content = multipart.getString(ServerMessageBuilder.MESSAGE_FIELD_MULTIPART_CONTENT);
        if (currentChunk == 1)
        {
            multipartMessageBuilder.setLength(0);
            multipartMessageBuilder.append(content);
            lastMultipartId = currentChunk;
        }
        else if (currentChunk != lastMultipartId + 1)
        {
            // TODO: LOG Multipart message received out of order. Discarding
            // message with multipart id %d.
        }
        else if (currentChunk == totalChunks)
        {
            multipartMessageBuilder.append(content);
            lastMultipartId = 0;
            onMessage(multipartMessageBuilder.toString());
        }
        else
        {
            multipartMessageBuilder.append(content);
            lastMultipartId = currentChunk;
        }
    }

    @Override
	public void onDisconnect(int code, String reason)
    {
    	Log.d(TAG, "onDisconnect: " + code + " - " + reason);

        switch (code)
        {
            case CLOSE_CANNOT_CONNECT:
                if (reachable && reconnectAttempt < RECONNECT_PERIODS.length)
                {
                    new AsyncTask<String, String, Boolean>()
                    {

                        @Override
                        protected Boolean doInBackground(String... params)
                        {
                            try
                            {
                                synchronized (ServerMessageTransport.this)
                                {
                                    if ((lookedUpAddresses == null) && (url != null))
                                    {
                                        host = new URI(url).getHost();
                                        lookedUpAddresses = InetAddress.getAllByName(host);
                                    }
                                }
                            }
                            catch (UnknownHostException e)
                            {
                                Log.e(TAG, "Could not look up host " + host);
                                if (listener != null)
                                {
                                    listener.onTransportFailed();
                                    return false;
                                }
                            }
                            catch (URISyntaxException e)
                            {
                                Log.e(TAG, "Could not parse hostname from URI " + url);
                                if (listener != null)
                                {
                                    listener.onTransportFailed();
                                    return false;
                                }
                            }
                            return true;
                        }

                        protected void onPostExecute(Boolean result)
                        {
                            if ((result) && (url != null) && reconnectAttempt < RECONNECT_PERIODS.length)
                            {
                            	final int addrIndex = reconnectAttempt % lookedUpAddresses.length; 
                            	final String address = lookedUpAddresses[addrIndex].getHostAddress();
                                scheduleReconnection(reconnectAttempt++, url.replaceFirst(host, address));
                            }
                        }
                    }.execute("");

                    return;
                }
                break;

            case CLOSE_DISCONNECTED:
                Log.i(TAG, "Disconnected by user: " + reason);
                // If this method is called due to a user initiated disconnect then we don't want to call
                // the transport failed call back, so return.
                return;
                
            default:
                if (reachable && reconnectAttempt < RECONNECT_PERIODS.length)
                {
                    scheduleReconnection(reconnectAttempt++, url);
                    return;
                }
        }

        if (listener != null)
        {
            // The WebSocket connection has been closed, this may cause a
            // problem if we have a call in progress.
            webSocket = null;
            listener.onTransportFailed();
        }
    }
    
    private int getReconnectPeriod(int attempt) {
    	// Return the specified index, or -1 if the specified value
    	// is greater than the array size. This should help with
    	// ArrayIndexOutOfBounds exceptions
    	if(attempt >= RECONNECT_PERIODS.length) {
    		return -1;
    	}
    	
    	return RECONNECT_PERIODS[attempt];
    }
    
    private void scheduleReconnection(final int attemptNo, final String url)
    {   
    	final int periodToWait = getReconnectPeriod(attemptNo);
    	
    	if(periodToWait == -1)
    	{
    		Log.i(TAG, "Connection failed, not performing reconnection attempt " + 
    	        attemptNo);
    		return;
    	}
    	
    	// Switch to next host and attempt reconnection
        Log.d(TAG, "Connection failed, performing reconnection attempt " + 
    	    attemptNo + " after " + periodToWait + "ms");
        
        if (listener != null)
        {
            listener.onTransportRetry(attemptNo, periodToWait);
        }
    	
    	new Handler(Looper.getMainLooper()).postDelayed(new Runnable() 
    	{	
            @Override
            public void run()
            {
            	new AsyncTask<Void, Void, Void>(){
					@Override
					protected Void doInBackground(Void... params) {
		                if (reachable)
		                {
    					    connectAttemptLock.lock();
    		                try
    		                {
    		                    if (webSocket != null)
    		                    {
    		                        webSocket.connect();
    		                    }
    		                    else
    		                    {
    		                        if (listener != null)
    		                        {
    		                            listener.onTransportFailed();
    		                        }
    		                    }
    		                }
    		                finally
    		                {
    		                    connectAttemptLock.unlock();
    		                }
		                }
						
						return null;
					}	
            	}.execute();
            }
        }, periodToWait);
    }

	@Override
	public void onMessage(byte[] data) {
		// TODO Auto-generated method stub
	}

	@Override
	public void onError(Exception error)
	{
	    if (!closing)
	    {
            // WEBRTC-7220  Log the exception, not just the message
            Log.e(TAG, "onError " + error, error);
    
            Log.w(TAG, "Exception when attempting to connect");
    
            onDisconnect(CLOSE_CANNOT_CONNECT, "Exception when attemping to connect");
	    }
	    else
	    {
	        onDisconnect(CLOSE_DISCONNECTED, error.getMessage());
	    }
	}

	public void setTrustManager(TrustManager trustManager) 	 
    { 	 
        this.trustManager = trustManager; 	 
    }
	
    public void setHostnameVerifier(HostnameVerifier hostnameVerifier)
    {
    	this.hostnameVerifier = hostnameVerifier;
    }
}
