/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.impl;

import org.json.JSONObject;

/**
 * A listener interface for the {@link ServerMessageTransport}. Implementations
 * will be informed when the transport has opened, when a message has been
 * received, and when the transport has failed.
 */
public interface ServerMessageTransportListener
{
    /**
     * Invoked when the transport opened.
     */
    void onTransportOpened();

    /**
     * Invoked when a server message has been received.
     * 
     * @param serverMessage The server message.
     */
    void onServerMessage(JSONObject serverMessage);

    /**
     * Invoked when the transport has failed and the number of retries has been
     * exhausted, or the transport legitimately closed but could not be reopened.
     */
    void onTransportFailed();
    
    /**
     * Invoked when the transport has failed, and an attempt to reestablish
     * has been sceheduled.
     * 
     * @param attemptNumber The current attempt number
     * @param delayUntilNextAttempt The time until the next attempt
     */
    void onTransportRetry(int attemptNumber, long delayUntilNextAttempt);
    
    /**
     * Invoked when the transport connection has been reestablished.
     */
    void onTransportReestablished();
}
