/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone;

import java.util.Set;

import com.alicecallsbob.fcsdk.android.ListenerTarget;

/**
 * Manages the audio device options presented by the hardware.
 * @author cafex
 *
 */
public interface AudioDeviceManager extends ListenerTarget<AudioDeviceManagerListener>
{
    enum AudioDevice
    {
        /** Speaker Phone */
        SPEAKER_PHONE,
        /** Wired handsfree headset */
        WIRED_HEADSET,
        /** The earpiece on a phone device */
        EARPIECE,
        /** Bluetooth headset with or without microphone */
        BLUETOOTH,
        /** Default */
        NONE
    }

    enum AudioManagerState
    {
        /** Not running */
        UNINITIALIZED,
        /** Running state */
        RUNNING
    }

    /** Trigger the checking of all connected audio devices */
    void updateAudioDeviceState();

    /**
     * Start the AudioDeviceManager service
     * 
     * @return true if the service started correctly
     */
    boolean start();

    /**
     * Stop the AudioDeviceManager service
     * 
     * @return true if the service stopped correctly
     */
    boolean stop();

    /**
     * Changes the default audio device. This will be used if there is no headset, wired or Bluetooth. Removing a wired
     * headset or powering off a Bluetooth device will cause fallback to this device.
     * 
     * @param defaultDevice The device to set as default, either Speaker or Earpiece
     */
    void setDefaultAudioDevice(AudioDevice defaultDevice);

    /**
     * Set the currently in use audio device. This will then trigger a refresh of all the devices's state. Only values
     * that have been returned by getAudioDevices() are valid.
     * 
     * @param defaultDevice The device to set as the current device.
     */
    void setAudioDevice(AudioDevice defaultDevice);

    /**
     * The currently in-use device
     * 
     * @return The device currently selected for Audio
     */
    AudioDevice getSelectedAudioDevice();

    /**
     * Get the list of currently connected and valid AudioDevices
     * 
     * @return The valid list
     */
    Set<AudioDevice> getAudioDevices();

}
