/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone;

import com.alicecallsbob.fcsdk.android.ListenerTarget;

/**
 * An object that represents a voice and/or video call.
 */
public interface Call extends ListenerTarget<CallListener>
{
    /**
     * Returns the call id.
     *
     * @return the call id.
     */
    String getCallId();

    /**
     * Returns the call status.
     *
     * @return the call status.
     */
    CallStatus getCallStatus();

    /**
     * Returns the remote address.
     *
     * @return the remote address.
     */
    String getRemoteAddress();

    /**
     * Returns the remote display name;
     *
     * @return the remote display name;
     */
    String getRemoteDisplayName();

    /**
     * Whether the local end of the call includes audio.
     *
     * @return <code>true</code> if the local call does have audio, <code>false</code> otherwise;
     */
    boolean hasLocalAudio();

    /**
     * Whether the local end of the call includes video.
     *
     * @return <code>true</code> if the local call does have video, <code>false</code> otherwise;
     */
    boolean hasLocalVideo();

    /**
     * Whether the remote SDP has audio.
     *
     * @return <code>true</code> if the remote SDP has audio, <code>false</code> otherwise;
     */
    boolean hasRemoteAudio();

    /**
     * Whether the remote SDP has video.
     *
     * @return <code>true</code> if the remote SDP has video, <code>false</code> otherwise;
     */
    boolean hasRemoteVideo();

    /**
     * Set the VideoSurface that renders the remote video stream.
     *
     * @param view the VideoSurface to render the remote video stream.
     */
    void setVideoView(VideoSurface view);

    /**
     * Answers an incoming call.
     * <p>
     * Note that the SDK only supports one call at a time. If you answer an incoming call
     * whilst another is in progress, then that other call will be terminated.
     *
     * @param audio Whether to use audio.
     * @param video Whether to use video.
     */
    void answer(boolean audio, boolean video);

    /**
     * Plays DTMF tones.
     *
     * @param code The DTMF tone(s) to play.
     * @param localPlayback Whether the tone sound should be echoed back to the user.
     */
    void playDTMFCode(String code, boolean localPlayback);

    /**
     * Terminates a call
     */
    void end();
    
    /**
     * Holds a call
     */
    void hold();
    
    /**
     * Resumes a call after holding
     */
    void resume();
    
    /**
     * Whether the remote SDP has audio.
     *
     * @return <code>true</code> if the remote SDP has audio, <code>false</code> otherwise;
     */
    boolean isHeld();
}
