/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone;

/**
 * A listener for the {@link Call} that will be informed of events.
 */
public interface CallListener
{
    /**
     * Callback indicating that an outgoing call failed to dial.
     * 
     * @param call The call that received the error.
     * @param message A descriptive message for the error.
     * @param callStatus The reason the call failed to dial.
     */
    void onDialFailed(Call call, String message, CallStatus callStatus);

    /**
     * Callback indicating that the an established call is reporting an error.
     * 
     * @param call The call that received the error.
     * @param message A descriptive message for the error.
     * @param callStatus The reason the call failed.
     */
    void onCallFailed(Call call, String message, CallStatus callStatus);
    
    /**
     * A callback notification indicating that a new SDP offer has been received
     * and has switched on or off audio and/or video streaming. Further action
     * is required when audio or video are re-enabled, i.e. deciding whether to
     * un-mute them, as they are automatically kept muted by the API.
     * 
     * When this method is called, the {@link Call#hasRemoteAudio()} and
     * {@link Call#hasRemoteVideo()} values will reflect the new values for
     * audio and video. The previous values for audio and video are passed in as
     * the hadAudio and hadVideo parameters of the method.
     * 
     * @param call The call that received the media change.
     * @param hadAudio The previous audio state before the media change.
     * @param hadVideo The previous video state before the media change.
     */
    void onMediaChangeRequested(Call call, boolean hadAudio, boolean hadVideo);

    /**
     * A callback notification indicates the call has changed state.
     * 
     * @param call The call that changed state.
     * @param status The new call state.
     */
    void onStatusChanged(Call call, CallStatus status);

    /**
     * A callback notification indicates the call has changed into an error state.
     * 
     * @param call The call that changed state.
     * @param statusInfo The new call state info.
     */
    void onStatusChanged(Call call, CallStatusInfo statusInfo);
    
    /**
     * A callback notification indicates the remote party display name has
     * changed.
     * 
     * @param call The call that changed.
     * @param name The new display name.
     */
    void onRemoteDisplayNameChanged(Call call, String name);

    /**
     * A callback notification indicating that the remote media stream was
     * added.
     * 
     * @param call The call that changed.
     */
    void onRemoteMediaStream(Call call);
    
    /**
     * A callback notification indicating that the perceived quality of the incoming audio 
     * and video streams has changed.
     * 
     * @param call The call that changed.
     * @param inboundQuality The quality of the stream(s) between 0 and 100, where 100 is best quality. 
     */
    void onInboundQualityChanged(Call call, int inboundQuality);
    
    /**
     * A callback notification indicates the remote party has placed us on hold (stopped receiving all media from us).
     * 
     * @param call The call that changed.
     */
    void onRemoteHeld(Call call);
    
    /**
     * A callback notification indicates the remote party has unheld us (resumed receiving all media from us).
     * 
     * @param call The call that changed.
     */
    void onRemoteUnheld(Call call);
}