/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone;

/**
 * Extended call status information 
 */
public class CallStatusInfo {

	private final CallStatus callStatus;
	private final boolean isError;
	private final String message;
	private final String reason;
	private final String errorCode;
	
	public CallStatusInfo(CallStatus callStatus) {
	    this(callStatus, false, null, null, null);	
	}
	
	protected CallStatusInfo(CallStatus callStatus, boolean isError, String message, String reason, String errorCode) {

		if(callStatus == null) {
			throw new RuntimeException("callStatus cannot be null");
		}

		this.callStatus = callStatus;
		this.isError = isError;
		this.message = message;
		this.reason = reason;
		this.errorCode = errorCode;
	}

	public CallStatus getCallStatus() {
		return callStatus;
	}

	public boolean isError() {
		return isError;
	}

	public String getMessage() {
		return message;
	}

	public String getReason() {
		return reason;
	}

	public String getErrorCode() {
		return errorCode;
	}
	
	@Override
	public String toString() {
		return String.format("%s callStatus:%s isError:%b message:%s reason:%s errorCode:%s", getClass().getSimpleName(), callStatus, isError, message, reason, errorCode);
	}

	/**
	 * Builder pattern factory for {@link CallStatusInfo} objects
	 */
	public static class Builder {
		
		private boolean isError = false;
		private String message;
		private String reason;
		private String errorCode;
		
		/**
		 * Create an instance of {@link CallStatusInfo}
		 * @param callStatus The callStatus to include. Must not be null.
		 * @return CallStatusInfo instance
		 */
		public CallStatusInfo build(CallStatus callStatus) {			
			return new CallStatusInfo(callStatus, isError, message, reason, errorCode);
		}
		
		/** true if this status represents an error */
		public Builder isError(boolean isError) {
			this.isError = isError;
			return this;
		}
		
		/** @param message Message detailing error */
		public Builder message(String message) {
			this.message = message;
			return this;
		}
		 
		/** @param reason Reason error occurred */
		public Builder reason(String reason) {
			this.reason = reason;
			return this;
		}

        /** @param errorCode errorIdentifier */
		public Builder errorCode(String errorCode) {
			this.errorCode = errorCode;
			return this;
		} 
	}
}
