/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone;

/**
 * Video capture resolutions that are supported by FCSDK.
 * 
 * @author CafeX Communications
 *
 */
public enum PhoneVideoCaptureResolution 
{
    /**
     * 174x144 
     * @deprecated Use RESOLUTION_176x144 instead
     */
    @Deprecated
    RESOLUTION_174x144(174, 144),

    /** 176x144 */
    RESOLUTION_176x144(176, 144),
    
    /** 352x288 */
    RESOLUTION_352x288(352, 288),
    
    /** 640x480 */
    RESOLUTION_640x480(640, 480),
    
    /** 960x720 */
    RESOLUTION_960x720(960, 720),
    
    /** 1280x720 */
    RESOLUTION_1280x720(1280, 720);
    
    /** The resolution width */
    private int width;
    
    /** The resolution height */
    private int height;
    
    PhoneVideoCaptureResolution(int width, int height)
    {
        this.width = width;
        this.height = height;
    }
    
    public Integer getWidth()
    {
        return width;
    }
    
    public Integer getHeight()
    {
        return height;
    }
    
    public String toString()
    {
        return width + "x" + height;
    }
    
    /**
     * Obtain a resolution enum from video dimensions
     * @param width The width
     * @param height The height
     * @return The corresponding video resolution
     * @throws ResolutionNotFoundException If the resolution is not found
     */
    public static PhoneVideoCaptureResolution getFromDimensions(int width, int height) 
            throws ResolutionNotFoundException
    {
        for (PhoneVideoCaptureResolution resolution : values())
        {
            if (resolution.width == width && resolution.height == height)
            {
                return resolution;
            }
        }
        
        throw new ResolutionNotFoundException(
                "Resolution: " + width + "x" + height + " not found.");
    }
}
