/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

public interface BluetoothDeviceManager
{

    enum State
    {
        /**
         * Bluetooth not available, or off
         */
        UNAVAILABLE,
        /**
         * Bluetooth proxy object for the Headset profile connected, connected Bluetooth headset present, but SCO is not
         * started or disconnected.
         */
        DEVICE_AVAILABLE,
        /**
         * Bluetooth proxy object for the Headset profile exists, but no connected headset devices, SCO is not started
         * or disconnected.
         */
        DEVICE_UNAVAILABLE,
        /**
         * Bluetooth audio SCO connection with remote device is initiated.
         */
        CONNECTING,
        /**
         * Bluetooth audio SCO connection with remote device is established.
         */
        CONNECTED,
        /**
         * Bluetooth audio SCO connection with remote device is closing.
         */
        DISCONNECTING;
    }

    /**
     * Start the service
     */
    void start();

    /**
     * Stop the service
     */
    void stop();

    /**
     * Return the current service state
     * @return current state
     */
    State getState();

    /**
     * Start a Bluetooth SCO conenction
     * @return true if it started
     */
    boolean startAudio();

    /**
     * Stop a Bluetooth SCO connection
     */
    void stopAudio();

    /**
     * Request a refresh the list of connected devices
     */
    void refreshListOfDevices();

    /**
     * Is a device currently connected
     * @return true if a device is paired, powered on and connected
     */
    boolean deviceIsConnected();

    /**
     * Is a device available to start an SCO connection.
     * @return True if a device is available
     */
    boolean deviceIsAvailable();
}
