/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ArrayBlockingQueue;

import org.acbrtc.MediaStream;
import org.acbrtc.StatsReport;

import android.media.AudioManager;
import android.media.ToneGenerator;
import android.text.TextUtils;
import android.util.Log;

import com.alicecallsbob.fcsdk.android.phone.Call;
import com.alicecallsbob.fcsdk.android.phone.CallListener;
import com.alicecallsbob.fcsdk.android.phone.CallStatus;
import com.alicecallsbob.fcsdk.android.phone.CallStatusInfo;
import com.alicecallsbob.fcsdk.android.phone.CallStatusInfo.Builder;
import com.alicecallsbob.fcsdk.android.phone.PhoneListener;
import com.alicecallsbob.fcsdk.android.phone.VideoSurface;

/**
 *
 */
public class CallImpl implements Call, CallStatisticsListener
{
	private static final String REASON_NO_MEDIA_BROKERS_AVAILABLE = "No Media Brokers available";
	
	private static final String REASON_REQUEST_TERMINATED = "Request Terminated";

	private static final String REASON_TEMPORARILY_UNAVAILABLE = "Temporarily Unavailable";

	protected static final String	TAG	= "CallImpl";

	protected static final int	DTMF_TONE_DURATION	= 200;	// 200 milliseconds

	// current status
	private CallStatus	mStatus = CallStatus.UNINITIALIZED;

	// audio media stream flag, does this call include audio?
	private boolean	mAudio = true;
	// video media stream flag, does this call include video?
	private boolean	mVideo = false;

	private boolean	mRemoteAudio = false, mRemoteVideo = false;

	private String	mID;

	private String	mRemoteAddress;
	private String	mRemoteDisplayName;
	   
    private boolean mEnding = false;
    
	public boolean	mIsCaller;
	
	private PeerConnectionClient mPeerConnClient;

	public enum Connectedness
	{
		UNKNOWN,
		DIALLING,
		RINGING,
		CONNECTED
	}
	private Connectedness	mConnectednessState = Connectedness.UNKNOWN;

	public enum SDPState
	{
        NONE,
        OFFERING,
		OFFERED,
		ANSWERED
	}
    
	private volatile SDPState mSDPState;
	private boolean mIceConnected = false;
	
	private boolean mRemoteHeld = false;
	private boolean mLocalHeld = false;

	private final PhoneImpl	mCallManager;
	private CallListener	mListener;

	private int mLastCallQuality;

	private final CallQualityCalculator callQualityCalculator;


	/**
	 * TimerTask for sending and playback of one or more DTMF tones, contained within a queue.
	 */
	private class DTMFTask extends TimerTask
	{
		private static final String NEXUS_7_ANDROID_OS_BUILD_DEVICE = "flo";

		private static final long	DTMF_PAUSE_DURATION	= (long)(DTMF_TONE_DURATION * 1.25); // 250ms

		private final ToneGenerator	mToneGenerator;
		private final boolean		mPlayback;

		public DTMFTask(boolean localPlayback)
		{
			mPlayback = localPlayback;
			mToneGenerator =
					new ToneGenerator(AudioManager.STREAM_DTMF, ToneGenerator.MAX_VOLUME);
		}

		@Override
		public void run()
		{			
		    // Save original microphone setting
	        final Map<String, Boolean> states = LocalMediaManager.getInstance().getLocalAudioEnabledState();
			
            String digit = null;
	        
		    while (!mDTMFQueue.isEmpty())
		    {		        
		        // Turn off microphone while playing tones
		        LocalMediaManager.getInstance().enableLocalAudio(false);
		        
                // Read the head of the queue into digit, checking to see if we
                // have emptied the queue (therefore null), a la ading byte
                // buffers from a stream
    			while ((digit = mDTMFQueue.poll()) != null )
    			{
    				// Play each digit with a 250ms pause to allow for the full tone to play
    				
    				/*
    				 * If this digit is sent/played, we pause before we send/play the next one, or
    				 * finish this task, so that each tone is played in full.
    				 */
    				if ( playDTMFDigit(digit) )
    				{
    					try
    					{
    						Thread.sleep(DTMF_PAUSE_DURATION);
    					}
    					catch (InterruptedException e)
    					{
    						e.printStackTrace();
    					}
    				}
    			}
    			
                // For the nexus 7, we need to allow a longer pause at the end
                // because some quirk of the device makes it particularly
                // susceptible to picking up DTMF tones

                if(NEXUS_7_ANDROID_OS_BUILD_DEVICE.equals(android.os.Build.DEVICE)) {
                    // At this point, the queue is empty, but another digit
                    // could come in while we're waiting. So rather than wait 
                    // for the additional delay to complete, interrupt it for
                    // the new digit
                    final long end = System.currentTimeMillis() + DTMF_PAUSE_DURATION;

                    do {
                        // Sleep for a bit
                        try {
                            Thread.sleep(10);
                        }
                        catch (InterruptedException e) {
                            e.printStackTrace();
                        }
                        // Until we've slept enough, or another DTMF code has come in
                    } while(System.currentTimeMillis() < end && mDTMFQueue.isEmpty()); 
                }
		    }

		    // Restore original microphone setting
			LocalMediaManager.getInstance().restoreLocalAudioEnabledState(states);

			mToneGenerator.release();

			mDTMFTimer.cancel();
			mDTMFTimer.purge();
			mDTMFTimer = null;
		}

		/**
		 * If the given digit is a valid DTMF code, send it to the server and play the tone locally
		 * if required.
		 *
		 *@param toneGen
		 * @param digit
		 * @param localPlayback
		 * @return true if a valid DTMF digit was specified and sent
		 */
		private boolean playDTMFDigit(String digit)
		{
			if ( digit.matches("[0-9A-D]|\\*|\\#") )
			{
				mCallManager.sendDTMFCode(CallImpl.this, digit);

				if ( mPlayback )
				{
					int	tone = ToneGenerator.TONE_DTMF_0;
					switch ( digit.charAt(0) )
					{
						case '0':	break;
						case '1':	tone = ToneGenerator.TONE_DTMF_1;	break;
						case '2':	tone = ToneGenerator.TONE_DTMF_2;	break;
						case '3':	tone = ToneGenerator.TONE_DTMF_3;	break;
						case '4':	tone = ToneGenerator.TONE_DTMF_4;	break;
						case '5':	tone = ToneGenerator.TONE_DTMF_5;	break;
						case '6':	tone = ToneGenerator.TONE_DTMF_6;	break;
						case '7':	tone = ToneGenerator.TONE_DTMF_7;	break;
						case '8':	tone = ToneGenerator.TONE_DTMF_8;	break;
						case '9':	tone = ToneGenerator.TONE_DTMF_9;	break;
						case '*':	tone = ToneGenerator.TONE_DTMF_S;	break;
						case '#':	tone = ToneGenerator.TONE_DTMF_P;	break;
						case 'A':	tone = ToneGenerator.TONE_DTMF_A;	break;
						case 'B':	tone = ToneGenerator.TONE_DTMF_B;	break;
						case 'C':	tone = ToneGenerator.TONE_DTMF_C;	break;
						case 'D':	tone = ToneGenerator.TONE_DTMF_D;	break;
					}

					// Play the tone for 200 milliseconds
					mToneGenerator.startTone(tone, DTMF_TONE_DURATION);
				}

				return true;
			}

			return false;
		}
	}

	/** Queue for playing DTMF tones */
	protected static ArrayBlockingQueue<String>	mDTMFQueue = new ArrayBlockingQueue<String>(20);
	protected static Timer	mDTMFTimer = null;

	/**
	 * Constructor
	 * @param callId
	 * @param address
	 * @param manager
	 */
	public CallImpl(String callId, String address, PhoneImpl manager, PeerConnectionClient peerConnClient)
	{
		mRemoteAddress = address;

		// if the call id was not defined, we are creating the call and should generate one
        if ( callId == null )
        {
            // http://www.ietf.org/rfc/rfc4122.txt
            char[]	s = new char[36];
            String	hexDigits = "0123456789abcdef";
            for ( int i = 0; i < 36; i++ )
                s[i] = hexDigits.charAt((int)Math.floor(Math.random() * 0x10));
            s[14] = '4'; // bits 12-15 of the time_hi_and_version field to 0100
            // bits 6-7 of the clock_seq_hi_and_reserved to 01
            s[19] = hexDigits.charAt((s[19] & 0x3) | 0x8);
            s[8] = s[13] = s[18] = s[23] = '-';

            callId = "CL-" + String.valueOf(s);
        }
        mID = callId;

		mCallManager = manager;
        setStatus(CallStatus.UNINITIALIZED);
        setSDPState(SDPState.NONE);
        mPeerConnClient = peerConnClient;
        callQualityCalculator = new CallQualityCalculator();
	}

	/**
	 *
	 * @param status
	 */
	void setStatus(CallStatus status)
	{
        Log.d(TAG, "Call status changing from " + mStatus + " to " + status);

		mStatus = status;
		if ( status == CallStatus.ENDED )
            onEnded();
	}

	/**
	 * Set the status of the SDP negotiation
	 * @param state
	 */
	void setSDPState(SDPState state)
	{
		mSDPState = state;
	}
	
    /**
     * Get the status of the SDP negotiation
     * @return state
     */
    SDPState getSDPState()
    {
        return mSDPState;
    }
   
    /**
     * Set the state of the Ice connection
     * @param connected
     */
    void setIceConnected(boolean connected)
    {
        mIceConnected = connected;
    }
   
    /**
     * Has the ice mechanism found a usable connection
     * @return true if a connection has been found
     */
    boolean isIceConnected()
    {
        return mIceConnected;
    }

	/**
	 *
	 * @param id
	 */
	public void setID(final String id)
	{
		mID = id;
	}

	@Override
	public String getRemoteAddress()
	{
		return mRemoteAddress;
	}

	/**
	 *
	 * @param address
	 */
	public void setRemoteAddress(final String address)
	{
		mRemoteAddress = address;
	}

	/**
	 *
	 * @param name
	 */
	public void setRemoteDisplayName(String name)
	{
		mRemoteDisplayName = name;
	}

	@Override
	public String getRemoteDisplayName()
	{
		return mRemoteDisplayName;
	}

	/**
	 *
	 * @param audio
	 * @param video
	 * @return boolean
	 */
	public boolean setup(boolean audio, boolean video)
	{
		Log.d(TAG, "setup");

		mAudio = audio;
		mVideo = video;

		if ( TextUtils.isEmpty(mRemoteAddress) )
		{
			Log.w(TAG, "remote address is null or empty");
			return false;
		}

		mIsCaller = true; // should only be true for outgoing calls

		if ( mCallManager.getLocalMediaStream() == null )
		{
			final String 	errorMsg = "Call setup failed: media has not been authorized yet";
			Log.e(TAG, errorMsg);
			if ( mListener != null )
				mListener.onCallFailed(this, errorMsg, CallStatus.ERROR);
			return false;
		}

	    setStatus(CallStatus.SETUP);
		return true;
	}

	@Override
	public void answer(boolean audio, boolean video)
	{
		Log.d(TAG, "answer");

		mAudio = audio;
		mVideo = video;

		mCallManager.processIncomingCall(this);
	}

	@Override
	public void end()
	{
		Log.d(TAG, "end");
		
		mEnding = true;
		
		mCallManager.endCall(this);
		
		setStatus(CallStatus.ENDED);
	}
	
	public void endOnDisconnect()
	{
	    Log.d(TAG, "endOnDisconnect");
	     
	    mEnding = true;
	            
	    mCallManager.endCall(this);
	        
	    setStatus(CallStatus.ENDED);
	}
	
	@Override
	public void hold()
	{
		Log.d(TAG, "hold");
		
		mLocalHeld = true;

		mCallManager.holdCall(this);
	}
	
	@Override
	public void resume()
	{
		Log.d(TAG, "resume");
		
		mLocalHeld = false;

		mCallManager.resumeCall(this);
	}

	/**
	 *
	 * @param audio
	 * @param video
	 * @return
	 */
	boolean remoteMediaChanged(boolean audio, boolean video)
	{
		Log.d(TAG, "mediaChanged, audio(" + audio + "), video(" + video + ")");

		mRemoteAudio = audio;
		mRemoteVideo = video;

		// If this is an outgoing audio-only call, we ignore incoming video
		if ( mIsCaller && (mVideo == false) && (video == true) )
			return false;

		if ( mListener != null )
			mListener.onMediaChangeRequested(this, mRemoteAudio, mRemoteVideo);

		// TODO: do we need to do anything else here?

		return true;
	}

	/**
	 * Is the call connectedness state RINGING?
	 * @return
	 */
	public boolean isRinging()
	{
		return (mConnectednessState == Connectedness.RINGING);
	}

	/**
	 * Is the call connectedness state DIALLING?
	 * @return
	 */
	public boolean isDialling()
	{
		return (mConnectednessState == Connectedness.DIALLING);
	}

	/**
	 * Is the call connectedness state CONNECTED?
	 * @return
	 */
	public boolean isConnected()
	{
		return (mConnectednessState == Connectedness.CONNECTED);
	}

	/**
	 *
	 * @param state
	 */
	public void setConnectedness(Connectedness state)
	{
		mConnectednessState = state;
	}

	/**
	 *
	 */
	public void updateStatusAndCallback()
	{
		switch ( mConnectednessState )
		{
			case CONNECTED:
			    Log.d(TAG, "updateStatusAndCallback. mConnectednessState: CONNECTED");
			    
			    boolean isSdpAnswered = getSDPState() == SDPState.ANSWERED;
			    boolean isIceConnected = isIceConnected();
			    boolean isInCall      = getCallStatus() == CallStatus.IN_CALL;
			    
			    Log.d(TAG, String.format("updateStatusAndCallback: isSdpAnswered:%b isIceConnected:%b isInCall:%b", isSdpAnswered, isIceConnected, isInCall));
			    
				if (isSdpAnswered && isIceConnected && !isInCall)
				{
	                onInCall();
				}
				else if (isSdpAnswered && !isIceConnected && !isInCall)
				{
					setStatus(CallStatus.MEDIA_PENDING);
					Log.i(TAG, "Pending: " + mRemoteAddress);
				}
				break;

			case RINGING:
			    Log.d(TAG, "updateStatusAndCallback. mConnectednessState: RINGING");
				if ( getCallStatus() != CallStatus.RINGING )
				{
				    setStatus(CallStatus.RINGING);
	        	    Log.i(TAG, "Ringing: " + mRemoteAddress);
	        	    if ( mListener != null )
	        			mListener.onStatusChanged(this, CallStatus.RINGING);
				}
				break;

			default: // UNKNOWN or DIALLING, we don't really care
				break;
		}
	}
	
	/**
	 *
	 */
	void onAlerting()
	{
	    Log.d(TAG, "Alerting: " + mRemoteAddress);
	    setStatus(CallStatus.ALERTING);

	    // Notify the CallListener
	    if ( mListener != null )
	    	mListener.onStatusChanged(this, CallStatus.ALERTING);

	    // Notify all the PhoneListeners
		if ( mCallManager.mListeners != null )
			for ( PhoneListener listener : mCallManager.mListeners )
				listener.onIncomingCall(this);
	}

	/**
	 * @param error Error identifier
	 * @param message Message detailing error
	 * @param reason Reason error occurred
	 */
	public void onError(final String error, final String message, final String reason)
	{
		Builder b = new CallStatusInfo.Builder();
		b.isError(true).message(message).reason(reason);
		
	    Log.e(TAG, error + " - " + message + " - " + reason);
	    setStatus(CallStatus.ERROR);
	    
	    // WEBRTC-7219 - Return a better status code when there is no MB capacity
	    if ( mListener != null ) {
	    
	    	if(reason == null) {
	    		mListener.onStatusChanged(this, b.build(CallStatus.ERROR));
	    	} else {
	    		if(reason.equalsIgnoreCase(REASON_NO_MEDIA_BROKERS_AVAILABLE)) {
		    		mListener.onStatusChanged(this, b.build(CallStatus.NO_MB_CAPACITY));
		    	} else if(reason.equalsIgnoreCase(REASON_REQUEST_TERMINATED)) {
		    		mListener.onStatusChanged(this, b.build(CallStatus.REQUEST_TERMINATED));
		    	} else if(reason.equalsIgnoreCase(REASON_TEMPORARILY_UNAVAILABLE)) {
		    		mListener.onStatusChanged(this, b.build(CallStatus.TEMPORARILY_UNAVAILABLE));
		    	} else {
				    mListener.onStatusChanged(this, b.build(CallStatus.ERROR));
		    	}	
	    	}
	    }
	}

    void onMediaUnavailable() {
        Log.e(TAG, "Media unavailable");
        setStatus(CallStatus.MEDIA_UNAVAILABLE);
        if ( mListener != null ) {
            mListener.onStatusChanged(this, CallStatus.MEDIA_UNAVAILABLE);
        }
    }

	/**
	 *
	 */
	private void onInCall()
	{
	    Log.d(TAG, "In-call: " + mRemoteAddress);
	    setStatus(CallStatus.IN_CALL);
	    if ( mListener != null )
			mListener.onStatusChanged(this, CallStatus.IN_CALL);
	}

	/**
	 *
	 */
	private void onEnded()
	{
	    Log.d(TAG, "Ended: " + mRemoteAddress);
	    if ( mListener != null )
			mListener.onStatusChanged(this, CallStatus.ENDED);
	}

	/**
	 *
	 */
	void onBusy()
	{
	    Log.d(TAG, "Busy: " + mRemoteAddress);
	    if ( mListener != null )
			mListener.onDialFailed(this, "The callee (" + mRemoteAddress + ") is busy", CallStatus.BUSY);
	}

	/**
	 *
	 */
	void onNotFound()
	{
	    Log.d(TAG, "Unreachable: " + mRemoteAddress);
	    if ( mListener != null )
			mListener.onDialFailed(this, "The callee (" + mRemoteAddress + ") is unreachable", CallStatus.NOT_FOUND);
	}

	/**
	 *
	 */
	public void onRemoteDisplayNameChanged()
	{
	    Log.d(TAG, "Remote display name updated: " + mRemoteDisplayName);
	    if ( mListener != null )
			mListener.onRemoteDisplayNameChanged(this, mRemoteDisplayName);
	}
	
	/**
    *
    */
   public void onRemoteHeld()
   {
       Log.d(TAG, "Remote held: " + mRemoteAddress);
       mRemoteHeld = true;
       if ( mListener != null )
           mListener.onRemoteHeld(this);
   }
   
   /**
   *
   */
  public void onRemoteUnheld()
  {
      Log.d(TAG, "Remote unheld: " + mRemoteAddress);
      mRemoteHeld = false;
      if ( mListener != null )
          mListener.onRemoteUnheld(this);
  }

	/**
	 *
	 */
	void onTimeout()
	{
	    Log.d(TAG, "Timeout: " + mRemoteAddress);
	    if ( mListener != null )
	    {
			mListener.onDialFailed(this, "The call (" + mRemoteAddress + ") timed out", CallStatus.TIMED_OUT);	    	
	    }
	}

	@Override
	public void addListener(CallListener listener)
	{
		mListener = listener;
	}

	@Override
	public void removeListener(CallListener listener)
	{
		mListener = null;
	}

	@Override
	public String getCallId()
	{
		return mID;
	}

	@Override
	public CallStatus getCallStatus()
	{
		return mStatus;
	}

	@Override
    public boolean hasLocalAudio()
    {
		return mAudio;
    }

	@Override
    public boolean hasLocalVideo()
    {
		return mVideo;
    }
	
	public void setHasLocalVideo(boolean hasLocalVideo)
	{
		mVideo = hasLocalVideo;
	}

	@Override
	public boolean hasRemoteAudio()
	{
		return mRemoteAudio;
	}

	@Override
	public boolean hasRemoteVideo()
	{
		return mRemoteVideo;
	}

	@Override
	public void setVideoView(VideoSurface view)
	{
		mPeerConnClient.setRemoteVideoView(view);
	}

	@Override
	public void playDTMFCode(final String code, final boolean localPlayback)
	{
		if ( code == null )
		{
			Log.e(TAG, "The DTMF code is null");
			return;
		}

        if(code.isEmpty()) {
            Log.i(TAG, "No DTMF code specified");
        }

		// Add each character of the code to the queue to be played
		int	numDigits = code.length();
		for ( int i=0; i<numDigits; i++ )
		{
			try
			{
				mDTMFQueue.add(code.substring(i, i + 1));
			}
			catch (IllegalStateException ise)
			{
				Log.w(TAG, "The DTMF queue is full, we can't add any more codes to the queue, so" +
						"wait for it to empty a bit.");
				return;
			}
		}

		if ( mDTMFTimer == null )
		{
			Log.v(TAG, "Schedule a new TimerTask to play the DTMF tone(s)");
			mDTMFTimer = new Timer("DTMF timer");
			mDTMFTimer.schedule(new DTMFTask(localPlayback), 0);
		}
	}

	/**
	 *
	 * @param stream
	 */
	void setRemoteMediaStream(final MediaStream stream)
	{
		if ( mListener != null )
			mListener.onRemoteMediaStream(this);
	}
	
	public boolean isLocalHeld()
	{
		return mLocalHeld;
	}
	
    boolean isEnding()
    {
        return mEnding;
    }
	
	public PeerConnectionClient getPeerConnectionClient()
	{
	    return mPeerConnClient;
	}

	@Override
	public void onCallStatisticsData(StatsReport[] remoteAudioStats, StatsReport[] remoteVideoStats) 
	{
		Log.i(TAG, "onCallStatisticsData " + remoteAudioStats + " - " + remoteVideoStats);
		
		final int callQuality = callQualityCalculator.calculateQuality(remoteAudioStats, remoteVideoStats);		 
		
		if (callQuality != mLastCallQuality)
		{
			mLastCallQuality = callQuality;

			if (mListener != null)
			{
			    mListener.onInboundQualityChanged(this, callQuality);
			}
		}
		else
		{
			Log.i(TAG, "onCallStatisticsData call quality un-changed " + mLastCallQuality);
		}
	}

    @Override
    public boolean isHeld()
    {
        return mRemoteHeld;
    }
}
