/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import android.media.AudioFormat;
import android.media.AudioRecord;
import android.media.MediaRecorder.AudioSource;
import android.util.Log;

/**
 * This checks whether the Microphone on the device is currently in use 
 *
 */
public final class MicrophoneAnalyser
{
	/** TAG */
	private static final String TAG = MicrophoneAnalyser.class.getSimpleName();

	/** Size of buffer to read data in to from the microphone */
	private static final int BYTE_BUFFER_SIZE = 20;

	private MicrophoneAnalyser()
	{
		
	}
	
	/**
	 * Checks if the Microphone is currently in use on the device.
	 * 
	 * @return true if it is, false if not
	 */
	public static boolean isMicrophoneInUse()
	{
		AudioRecord audioRecord = null;
		try
		{
			audioRecord = createAudioRecorderForMicrophone();
			audioRecord.startRecording();

			final byte[] buffer = new byte[BYTE_BUFFER_SIZE];
			final int returnVal = audioRecord.read(buffer, 0, BYTE_BUFFER_SIZE);

			if (returnVal == AudioRecord.ERROR_BAD_VALUE)
			{
				Log.d(TAG, "Got ERROR_BAD_VALUE whilst recording from the Mic - Microphone may be in use");
				return true;
			}
			if (returnVal == AudioRecord.ERROR_INVALID_OPERATION)
			{
				Log.d(TAG, "Got ERROR_INVALID_OPERATION whilst recording from the Mic - Microphone may be in use");
				return true;

			}	
			audioRecord.stop();		
		}
		catch (Exception e)
		{
			Log.e(TAG, "Error whilst checking if the Mic is in use", e);
		}
		finally
		{
			if (audioRecord != null)
			{
				audioRecord.release();
			}
		}

		return false;
	}

	private static AudioRecord createAudioRecorderForMicrophone()
	{
		final int sampleRateInHz = 8000;
		final int channelConfig = AudioFormat.CHANNEL_IN_MONO;
		final int audioFormat = AudioFormat.ENCODING_PCM_16BIT;
		final int bufferSize = AudioRecord.getMinBufferSize(sampleRateInHz,
		        channelConfig, audioFormat);

		final AudioRecord audioRecord = new AudioRecord(AudioSource.MIC, 8000,
		        AudioFormat.CHANNEL_IN_MONO, AudioFormat.ENCODING_PCM_16BIT,
		        bufferSize);

		return audioRecord;
	}
}
