/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import org.acbrtc.DataChannel;
import org.acbrtc.IceCandidate;
import org.acbrtc.MediaConstraints;
import org.acbrtc.MediaConstraints.KeyValuePair;
import org.acbrtc.MediaStream;
import org.acbrtc.PeerConnection;
import org.acbrtc.PeerConnection.IceConnectionState;
import org.acbrtc.PeerConnection.IceGatheringState;
import org.acbrtc.PeerConnection.IceServer;
import org.acbrtc.PeerConnection.SignalingState;
import org.acbrtc.PeerConnectionFactory;
import org.acbrtc.SdpObserver;
import org.acbrtc.SessionDescription;
import org.acbrtc.VideoRenderer;
import org.acbrtc.VideoTrack;

import android.content.Context;
import android.util.Log;

import com.alicecallsbob.fcsdk.android.phone.VideoSurface;

/**
 * TODO: Implement one peer connection per call.
 */
class PeerConnectionClient implements PeerConnection.Observer
{
	private static final String	TAG = "PeerConnectionClient";

	private static final String	MEDIA_STREAM_LABEL_LOCAL	= "local_media_stream";
	private static final String	MEDIA_STREAM_LABEL_REMOTE	= "remote_media_stream";

	// Used along the same lines as the error types in CallServerMessageBuilder
	public static final String MESSAGE_ERROR_MEDIA_UNAVAILABLE = "MEDIA_UNAVAILABLE";

	private PeerConnectionFactory	mFactory;
	private PeerConnection			mPeerConnection;
	private LocalMediaManager       mLocalMediaManager;
	private CallStatisticsMonitor 	mCallStatisticsMonitor;
	
	public boolean	candidatesComplete;

	private PhoneImpl	mCallManager;
	
	private CallImpl mCall;

	private Hashtable<String, MediaStream>	mActiveStreams = null;

	private VideoSurfaceImpl	mRemoteVideoView;
	private int					mCameraIndex;
	private VideoRenderer       mRemoteRenderer;

	private Context mContext;

    private VideoSurfaceImpl.VideoRendererCallbacks remoteCallbacks;

	
	/**
	 * 
	 * @param context
	 * @param callManager
	 */
	public PeerConnectionClient(Context context, PhoneImpl callManager)
	{
		Log.d(TAG, "Constructor");
		
		mCallManager = callManager;

		try
		{
			mFactory = PeerConnectionFactoryManager.getInstance().getFactory(context);
		}
		catch (RuntimeException re)
		{
			Log.e(TAG, "Failed to get PeerConnectionFactory. Error Details: " + re);
			re.printStackTrace();
			return;
		}

		if ( mFactory == null )
		{
			Log.e(TAG, "Failed to initialize PeerConnectionFactory");
			return;
		}

		Log.d(TAG, "PeerConnectionFactory creation success");
		
		mLocalMediaManager = LocalMediaManager.getInstance();		
		mCameraIndex = mCallManager.getCamera();
		mActiveStreams = new Hashtable<String, MediaStream>();
		mContext = context;
	}

	/**
	 * @param context
	 * @param callManager
	 * @return
	 */
	public static PeerConnectionClient create(Context context, PhoneImpl callManager)
	{
		Log.d(TAG, "create");

		return new PeerConnectionClient(context, callManager);
	}

	/**
	 *
	 */
	public void destroy()
	{
		Log.d(TAG, "destroy");

		closePeerConnection();
	}

	/**
	 *
	 * @param remoteView
	 */
	void setRemoteVideoView(VideoSurface remoteView)
	{
		Log.d(TAG, "setRemoteVideoView " + mPeerConnection);
		
		mRemoteVideoView = (VideoSurfaceImpl)remoteView;

		final MediaStream stream = mActiveStreams.get(MEDIA_STREAM_LABEL_REMOTE);
		
		if (remoteView != null)
		{    
    		// If we're in-call, we need to set this view as the renderer for the video track
    		if ( mPeerConnection != null )
    		{
    			Log.d(TAG, "setRemoteVideoView stream " + stream);
    
    			if ( stream != null )
    				addRemoteVideoRenderer(stream);
    		}
		}
		else
		{
		    mRemoteVideoView = null;
		    removeRemoteVideoRenderer(stream);
		}
	}

	/**
	 *
	 * @param localView
	 */
	void setLocalVideoView(VideoSurface localView)
	{
		Log.d(TAG, "setLocalVideoView " + mPeerConnection);
		
		mLocalMediaManager.setLocalVideoView(localView);
	}

	/**
	 * @param video - Is the current/new call a video call? <b>true</b> if so, <b>false</b>
	 * otherwise
	 * @return
	 */
	public synchronized boolean createPeerConnection(boolean video, CallImpl call, List<IceServer> iceServers)
	{
		Log.d(TAG, "createPeerConnection " + iceServers);

		if ( mPeerConnection != null )
		{
			Log.v(TAG, "A peer connection already exists");
			return true;
		}

		candidatesComplete = true;
		
		if (iceServers == null)
		{
			iceServers = new Vector<IceServer>();
		}
		
		if (iceServers.isEmpty())
		{
			IceServer server = new IceServer(/*"stun:stun.l.google.com:19302"*/"");
			iceServers.add(server);
		}

		MediaConstraints constraints = new MediaConstraints();
		KeyValuePair enableDtls = new KeyValuePair("DtlsSrtpKeyAgreement", "true");
		constraints.optional.add(enableDtls);
		
		mPeerConnection = mFactory.createPeerConnection(iceServers, constraints, this);
		if ( mPeerConnection == null )
		{
			Log.e(TAG, "createPeerConnection failed");
			return false;
		}

		MediaStream stream;
		
		if (video)
		{
		    stream = mLocalMediaManager.getOrCreateLocalVideoStream(mFactory, mCameraIndex, mCallManager);
		    
		    if(stream == null || stream.videoTracks == null || stream.videoTracks.isEmpty() ) {
		    	Log.e(TAG, "createPeerConnection create video stream failed");
				mCallManager.handleError(call.getCallId(), MESSAGE_ERROR_MEDIA_UNAVAILABLE, "Local video unavailable");
				return false;
		    }
		}
		else
		{
		    stream = mLocalMediaManager.getOrCreateLocalAudioStream(mFactory);
		    
		    if(stream == null || stream.audioTracks == null || stream.audioTracks.isEmpty() ) {
		    	Log.e(TAG, "createPeerConnection create audio stream failed");
				mCallManager.handleError(call.getCallId(), MESSAGE_ERROR_MEDIA_UNAVAILABLE, "Local audio unavailable");
				return false;
		    }
		}
				
		if ( mPeerConnection.addStream(stream, new MediaConstraints()) )
		{
			mActiveStreams.put(MEDIA_STREAM_LABEL_LOCAL, stream);

			if (call == null)
			{
			    mCall = mCallManager.getCurrentCallImpl();
			}
			else
			{
			    mCall = call;
			}
			
			mCallManager.setLocalMediaStream(stream);
		}
		else
			Log.e(TAG, "Adding local media stream to peer connection failed");

		return true;
	}
	
	private void removeRemoteVideoRenderer(MediaStream stream)
	{
	    if (stream != null)
	    {
    	    VideoTrack vt = stream.videoTracks.get(0);
            if ( vt != null && mRemoteRenderer != null)
            {            
                Log.d(TAG, "Removing renderer: " + mRemoteRenderer);
                vt.removeRenderer(mRemoteRenderer);
                mRemoteRenderer = null;
            }
	    }
	}

	/**
	 * Add the remote video view as the renderer for the given MediaStream.
	 * @param stream
	 */
	private void addRemoteVideoRenderer(MediaStream stream)
	{
		Log.d(TAG, "addRemoteVideoRenderer");

		if ( mRemoteVideoView == null )
		{
			Log.d(TAG, "UI remote video view is null");
		}
		else if (stream.videoTracks != null && stream.videoTracks.size() > 0)
		{
			Log.d(TAG, "Adding REMOTE video renderer " + stream.videoTracks.size());

			VideoTrack	vt = stream.videoTracks.get(0);
            if ( vt != null )
			{
                removeRemoteVideoRenderer(stream);
                
				remoteCallbacks = mRemoteVideoView.createCallbacks(VideoSurface.Endpoint.REMOTE);				
				mRemoteRenderer = new VideoRenderer(remoteCallbacks);
								
				Log.d(TAG, "Adding REMOTE video renderer");
				vt.addRenderer(mRemoteRenderer);

			}
            else
            {
				Log.w(TAG, "the video track is null");
            }
		}
	}

	/**
	 *
	 */
	public synchronized void closePeerConnection()
	{
		Log.d(TAG, "closePeerConnection");

		// AV-373 - explicitly destroy the callback to break references
        if(remoteCallbacks != null) {
            remoteCallbacks.destroy();
        }

        if(mRemoteVideoView != null) {
            mRemoteVideoView.destroy();
        }
		
        /*
         * We need to stop and dispose of the video capturer so that the Camera
         * resource is released and we're able to make another call without
         * restarting the app.
         */
        if (mPeerConnection != null)
        {   
            mRemoteRenderer = null;
			if ( mCall != null )
			{
				mCall.setRemoteMediaStream(null);
			}

			MediaStream localStream = mActiveStreams.get(MEDIA_STREAM_LABEL_LOCAL);
    		if ( mActiveStreams != null )
    		{
    			mActiveStreams.clear();
    			//mActiveStreams = null; Do not remove as reused for each call.
    		}
        	
    		mRemoteVideoView = null;

    		/*
        	 * The PeerConnection.dispose() method recursively cleans up child resources. It works something like this:
        	 * 
        	 * 1) PeerConnection.dispose() calls PeerConnection.close()
        	 * 2) PeerConnection.dispose() will remove each MediaStream and call MediaStream.dispose()
        	 * 3) MediaStream.dispose() will remove each AudioTrack and call AudioTrack.dispose()
        	 * 4) MediaStream.dispose() will remove each VideoTrack and call VideoTrack.dispose()
        	 * 5) VideoTrack.dispose() will remove each VideoRenderer and call VideoRenderer.dispose()
        	 * 
        	 * Previous versions of this method attempted to explicitly perform the above cleanup operations,
        	 * however this proved unreliable, it would invariable cause intermittent segmentation faults. Best 
        	 * not to interfere with the cleanup. 
        	 */


    		final boolean localStillUsed = mCallManager.getCurrentCalls().size() > 1;
            
            if( mCallStatisticsMonitor != null)
            {
                mCallStatisticsMonitor.stop();
                mCallStatisticsMonitor = null;
            }
            
    		if (!localStillUsed)
    		{  		  
    		    localStream = null;
    		    mCallManager.setLocalMediaStream(null);
        		mPeerConnection.dispose();      		
        		mPeerConnection = null;
                
                mLocalMediaManager.disposeVideoCapturer();
                mLocalMediaManager = null;
    		}
    		else
    		{
    		    Log.i(TAG, "There is still an ongoing call");
    		    //If this is not the only call, then the local stream is still in use so we should
    		    //remove it from the peer connection before disposing, so other calls remain active
    		    mPeerConnection.removeStream(localStream);
    		    mPeerConnection.dispose();
    		    mPeerConnection = null;
     		}    		
        }
	}

	/**
	 * @see org.acbrtc.PeerConnection.Observer#onSignalingChange(org.acbrtc.PeerConnection.SignalingState)
	 */
	@Override
	public void onSignalingChange(SignalingState sigState)
	{
		Log.v(TAG, "onSignalingState: " + sigState.name());
	}

	/**
	 * @see org.acbrtc.PeerConnection.Observer#onIceConnectionChange(org.acbrtc.PeerConnection.IceConnectionState)
	 */
	@Override
	public void onIceConnectionChange(IceConnectionState iceConnState)
	{
		Log.v(TAG, "onIceConnectionChange: " + iceConnState.name());
		if ( iceConnState == IceConnectionState.CONNECTED )
		{
			if ( mCall != null )
			{
		        mCall.setIceConnected(true);
		        mCall.updateStatusAndCallback();
			}
		}
	}

	/**
	 * @see org.acbrtc.PeerConnection.Observer#onIceGatheringChange(org.acbrtc.PeerConnection.IceGatheringState)
	 */
	@Override
	public void onIceGatheringChange(IceGatheringState state)
	{
		Log.d(TAG, "onIceGatheringChange: " + state.name());
		if ( state == IceGatheringState.COMPLETE )
			candidatesComplete = true;
	}

	/**
	 * @see org.acbrtc.PeerConnection.Observer#onIceCandidate(org.acbrtc.IceCandidate)
	 */
	@Override
	public void onIceCandidate(final IceCandidate candidate)
	{
		Log.v(TAG, "onIceCandidate recieved " + candidate);

		if (candidate != null)
		{
			mCallManager.onIceCandidate(mCall, candidate);
		}
		
		// No need to do anything for standard strategy as Media Broker knows the client's address.
	}

	public void onRemoteIceCandidate(final IceCandidate candidate)
	{
		if (candidate != null && mPeerConnection != null)
		{
			mPeerConnection.addIceCandidate(candidate);
		}
	}

	/**
	 * @see org.acbrtc.PeerConnection.Observer#onError()
	 */
	@Override
	public void onError()
	{
		Log.d(TAG, "onError");
		mCallManager.handleError(null, null, "PeerConnection error!");
	}

	/**
	 * Triggered when media is received on a new stream from a remote peer.
	 * @see org.acbrtc.PeerConnection.Observer#onAddStream(org.acbrtc.MediaStream)
	 */
	@Override
	public void onAddStream(final MediaStream stream)
	{
		Log.d(TAG, "onAddStream: " + stream.toString());

		final boolean	remoteAudio = (stream.audioTracks.size() == 1);
		final boolean	remoteVideo = (stream.videoTracks.size() == 1);

		Log.d(TAG, "onAddStream 2: " + mCall);
		if (mCall != null)
		{
			boolean	bSetupRemoteVideo = true;

			if (	(remoteAudio != mCall.hasRemoteAudio())
				||	(remoteVideo != mCall.hasRemoteVideo()) )
			{
				bSetupRemoteVideo = mCall.remoteMediaChanged(remoteAudio, remoteVideo);
			}

			if ( remoteVideo && bSetupRemoteVideo )
			{
				Log.v(TAG, "Got a new remote video stream.");
				addRemoteVideoRenderer(stream);
			}

//	        /*
//	         * If the new stream does not have video, but we have set up local
//	         * video, we need to clean up the local video components.
//	         */
//	        if (!remoteVideo && (mVideoCapturer != null))
//	        {
//	            mVideoCapturer.stop();
//	            mVideoCapturer.dispose();
//	            mVideoCapturer = null;
//	        }

	        mActiveStreams.put(MEDIA_STREAM_LABEL_REMOTE, stream);
	        mCall.setRemoteMediaStream(stream);
	        
	        // Monitor remote stream statistics.
			if (mCallStatisticsMonitor == null)
			{
				mCallStatisticsMonitor = new CallStatisticsMonitor(mPeerConnection, mActiveStreams, mCall, mContext);
				mCallStatisticsMonitor.start();
			}
			else
			{
				Log.i(TAG, "Using existing CallStatisticsMonitor after new remote stream added.");
			}
		}
	}

	/**
	 * Triggered when a remote peer closes a stream.
	 * @see org.acbrtc.PeerConnection.Observer#onRemoveStream(org.acbrtc.MediaStream)
	 */
	@Override
	public void onRemoveStream(final MediaStream stream)
	{
		Log.v(TAG, "onRemoveStream: " + stream.toString());
		stream.videoTracks.get(0).dispose();
	}

	/**
	 *
	 * @param mc
	 */
	public void createOffer()
	{
		createOffer(null);
	}
	
	/**
	 * 
	 * @param observer
	 */
	private synchronized void internalCreateOffer(SdpObserver observer)
	{
        if ( mPeerConnection != null )
        {
            MediaConstraints    constraints = new MediaConstraints();
            constraints.mandatory.add(new MediaConstraints.KeyValuePair("OfferToReceiveAudio",
                                                                        "true"));
            constraints.mandatory.add(new MediaConstraints.KeyValuePair("OfferToReceiveVideo",
                                                                        "true"));
            mPeerConnection.createOffer(observer, constraints);
        }
	}
	
	/**
	 * 
	 * @param observer
	 */
	public void createOffer(SdpObserver observer)
	{
	    Log.d(TAG, "createOffer");
        internalCreateOffer(new SdpObserverWrapper(observer));
	}
	
	/**
	 * 
	 * @param observer
	 */
	public void createHoldOffer(SdpObserver observer)
	{
	    Log.d(TAG, "createHoldOffer");
        internalCreateOffer(new SdpObserverWrapperForHold(observer));
	}

	/**
	 *
	 * @return
	 */
	public String getLocalDescription()
	{
		if ( mPeerConnection != null )
		{
			SessionDescription	sd = mPeerConnection.getLocalDescription();
			if ( sd != null )
				return sd.description;
		}

		return null;
	}
	
	/**
	 * Get the remote description.
	 * 
	 * @return The remote SDP.
	 */
	public String getRemoteDescription()
	{
	    if ( mPeerConnection != null )
        {
            SessionDescription  sd = mPeerConnection.getRemoteDescription();
            if ( sd != null )
                return sd.description;
        }

        return null;
	}

	/**
	 *
	 * @param type
	 * @param sdp
	 */
	public void setRemoteDescription(final String type, final String sdp)
	{
		Log.d(TAG, "setRemoteDescription");

		if ( mPeerConnection == null )
		{
			Log.w(TAG, "Cannot set the remote description on a null PeerConnection");
			return;
		}

        SessionDescription	description = null;

        if ( type != null )
    		description = new SessionDescription(SessionDescription.Type.fromCanonicalForm(type),
    		                                     sdp);

        mPeerConnection.setRemoteDescription(new SdpObserverWrapper(null), description);
	}

	/**
	 *
	 */
	public void createAnswer(SdpObserver observer)
	{
		Log.d(TAG, "createAnswer");
		mPeerConnection.createAnswer(new SdpObserverWrapper(observer), new MediaConstraints());
	}
	
	/**
    *
    */
   public void createForcedHoldAnswer(SdpObserver observer)
   {
       Log.d(TAG, "createForcedHoldAnswer");
       mPeerConnection.createAnswer(new SdpObserverWrapperForHold(observer), new MediaConstraints());
   }

	/**
	 *
	 * @param index
	 */
	public void setCameraIndex(int index)
	{
		mCameraIndex = index;
	}

	@Override
	public void onDataChannel(DataChannel arg0)
	{
	}
	
	/**
	 * Wraps an observer so that the local description is always set when
	 * SDP is created successfully.
	 * 
	 * @author CafeX Communications
	 *
	 */
	class SdpObserverWrapper implements SdpObserver
	{
	    protected SdpObserver observer;
	    
	    public SdpObserverWrapper(SdpObserver observer)
	    {
	        this.observer = observer;
	    }

        @Override
        public void onCreateFailure(String arg0) 
        {
            Log.w(TAG, "Failed to create SDP: " + arg0);
            if (observer != null)
            {
                observer.onCreateFailure(arg0);
            }          
        }

        @Override
        public void onCreateSuccess(SessionDescription sdp) 
        {
            // Note that setting the local description triggers ICE candidate detection.
        	
        	mPeerConnection.setLocalDescription(this, sdp);
        	
            if (observer != null)
            {
                observer.onCreateSuccess(sdp);
            }          
        }

        @Override
        public void onSetFailure(String arg0) 
        {
            Log.w(TAG, "Failed to set SDP: " + arg0);
            if (observer != null)
            {
                observer.onSetFailure(arg0);
            }
            
        }

        @Override
        public void onSetSuccess() 
        {
            if (observer != null)
            {
                observer.onSetSuccess();
            }
            
        }
	}
	
	/**
	 * Wraps an observer to convert created SDP into hold SDP.
	 * 
	 * @author CafeX Communications
	 *
	 */
	class SdpObserverWrapperForHold extends SdpObserverWrapper
	{
        public SdpObserverWrapperForHold(SdpObserver observer) 
        {
            super(observer);
        }
        
        @Override
        public void onCreateSuccess(SessionDescription sdp) 
        {
            final String description = sdp.description;
            
            String holdSDP = description.replace("sendrecv", "inactive");
            holdSDP = holdSDP.replace("sendonly", "inactive");
            holdSDP = holdSDP.replace("recvonly", "inactive");
            
            mPeerConnection.setLocalDescription(this, new SessionDescription(sdp.type, holdSDP));
            
            if (observer != null)
            {
                observer.onCreateSuccess(sdp);
            }
        }	    
	}
	
	public void setLocalDescription()
	{
		Log.d(TAG, "setLocalDescription being called !!!");
	}
}
