/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.phone.impl;

import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import org.acbrtc.IceCandidate;
import org.acbrtc.MediaStream;
import org.acbrtc.PeerConnection.IceServer;
import org.acbrtc.SdpObserver;
import org.acbrtc.SessionDescription;
import org.acbrtc.videoengine.VideoCaptureAndroid;

import com.alicecallsbob.fcsdk.android.impl.ServerMessageListenerTargetImpl;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageManager;
import com.alicecallsbob.fcsdk.android.phone.AudioDeviceManager;
import com.alicecallsbob.fcsdk.android.phone.Call;
import com.alicecallsbob.fcsdk.android.phone.CallCreationError;
import com.alicecallsbob.fcsdk.android.phone.CallCreationWithErrorException;
import com.alicecallsbob.fcsdk.android.phone.CallListener;
import com.alicecallsbob.fcsdk.android.phone.CallStatus;
import com.alicecallsbob.fcsdk.android.phone.Phone;
import com.alicecallsbob.fcsdk.android.phone.PhoneListener;
import com.alicecallsbob.fcsdk.android.phone.PhoneVideoCaptureResolution;
import com.alicecallsbob.fcsdk.android.phone.PhoneVideoCaptureSetting;
import com.alicecallsbob.fcsdk.android.phone.VideoSurface;
import com.alicecallsbob.fcsdk.android.phone.VideoSurfaceListener;
import com.alicecallsbob.fcsdk.android.phone.impl.CallImpl.SDPState;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.AnswerMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.CallConfigMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.CandidateMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.DisplayNameMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.EndMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.ErrorMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.EstablishedMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.OfferMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.OfferRequestMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.PingMessageHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.RemoteHeldHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.RemoteUnheldHandler;
import com.alicecallsbob.fcsdk.android.phone.impl.handlers.RingingMessageHandler;

import android.bluetooth.BluetoothAdapter;
import android.content.Context;
import android.graphics.Point;
import android.text.TextUtils;
import android.util.Log;
import junit.framework.Assert;

/**
 */
public class PhoneImpl extends ServerMessageListenerTargetImpl<PhoneListener> implements Phone
{
    /** The height of a 720p HD resolution */
    private static final int SEVEN_TWENTY_RESOLUTION_HEIGHT = 720;
    
    /** The minimum number of cores required to support 720p or higher */
    private static final int MIN_CORES_FOR_720 = 2;
    
	protected static final String	TAG = "PhoneImpl";
	
	//Supported framerates in android are multiplied by 1000
	protected static final int ANDROID_FRAMERATE_MULTIPLIER = 1000;

	protected static final int SUGGESTED_FRAMERATE = 30;
	
	/** Number of milliseconds in a second */
	private static final int SECOND_IN_MS = 1000;
	
	/** Number of times we check for received call config per second on call setup */
	private static final int CALL_CONFIG_RETRIES_PER_SECOND = 4;

	// the current list of calls
	private final ArrayList<Call>	mCurrentCalls = new ArrayList<Call>();

	public final ArrayList<PhoneListener>	mListeners;

	private volatile boolean mOutboundOfferPending;
	
	private int mCamera;

	private Timer mIceCandidateTimer = new Timer();

	private Timer trickleIceTimer = new Timer();
	
	/** List of Ice Servers recieved in call config, keyed on call ID */
	private Map<String, List<IceServer>> iceServers = new HashMap<String, List<IceServer>>();
	
	private volatile int swiftVersion = 0;
	private boolean trickleIceStrategy = false;
	
	private CameraVideoAnalyser cameraAnalyser;
		
	/**
	 *
	 */
	private interface SetupHandler
	{
		void handleSetup(CallImpl call);
	}

	protected String	mQueuedRemoteNameChange;

	protected Timer	mGeneralErrorTimer;

	protected VideoSurface	mVideoView;
	protected VideoSurface	mPreviewView;
	protected PhoneVideoCaptureResolution mPreferredCaptureResolution = 
	        PhoneVideoCaptureResolution.RESOLUTION_640x480;
	protected int mPreferredFramerate = 30;
	protected Context mContext;
	protected int mOrientation = 0;
	protected MediaStream mLocalMediaStream;

	protected AudioDeviceManager mAudioDeviceManager;
	protected BluetoothAdapter mBluetoothAdapter;
	
	/**
	 * Constructor
	 * @param context
	 * @param serverMessageManager
	 */
	public PhoneImpl(Context context, ServerMessageManager serverMessageManager)
	{
		super(serverMessageManager);
		Log.d(TAG, "Constructor");

		mListeners = new ArrayList<PhoneListener>();
		
		mContext = context;

		mQueuedRemoteNameChange = null;
		
		trickleIceStrategy = false;

		setupHandlers();

		cameraAnalyser = new CameraVideoAnalyser();
		
		mCamera = cameraAnalyser.getDefaultCameraIdxToUse();
		cameraAnalyser.setSupportedCamerasSettingsAndFramerates();
		
		mAudioDeviceManager = AudioDeviceManagerImpl.create(context);
		Log.d(TAG, "Starting the audio manager...");
	    mAudioDeviceManager.start();
	}
	
	@Override
	public VideoSurface createVideoSurface(Context c, Point screenDimensions, VideoSurfaceListener listener)
	{
		return new VideoSurfaceImpl(c, screenDimensions, listener);
	}
	
	/**
	 *
	 */
	private void setupHandlers()
	{
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_ANSWER, new AnswerMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_OFFER, new OfferMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_OFFER_REQUEST,
		           new OfferRequestMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_DISPLAY_NAME,
		           new DisplayNameMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_END, new EndMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_ERROR, new ErrorMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_RINGING, new RingingMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_ESTABLISHED,
		           new EstablishedMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_PING, new PingMessageHandler(this));

		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_CANDIDATE, new CandidateMessageHandler(this));
		addHandler(CallServerMessageBuilder.MESSAGE_TYPE_CALL_CONFIG, new CallConfigMessageHandler(this));
        addHandler(CallServerMessageBuilder.MESSAGE_TYPE_REMOTE_HELD, new RemoteHeldHandler(this));
        addHandler(CallServerMessageBuilder.MESSAGE_TYPE_REMOTE_UNHELD, new RemoteUnheldHandler(this));
	}

	/**
	 *
	 */
	public void destroy()
	{
		Log.d(TAG, "destroy");

		mQueuedRemoteNameChange = null;
		
        if ( iceServers != null )
        {
            iceServers.clear();
        }

		for (Call call : mCurrentCalls)
		{
		   ((CallImpl)call).getPeerConnectionClient().destroy();
		}
		clearCalls();
		
		mAudioDeviceManager.stop();
	}

	/**
	 *
	 */
	public void clearCalls()
	{
		if ( mCurrentCalls != null )
			mCurrentCalls.clear();
	}

	private CallImpl createNewCall(String address, boolean addVideo) {
		return createNewCall(address, addVideo, null);
	}
	
	/**
	 * @param address
	 * @return
	 */
	private CallImpl createNewCall(String address, boolean addVideo, CallListener listener)
	{
		Log.d(TAG, "create a new Call object cm = " + this);

		CallImpl	call = new CallImpl(null, address, this, new PeerConnectionClient(mContext, this)); // pass null for callId 'cos it doesn't exist yet
		if(listener != null) {
			call.addListener(listener);
		}
		mCurrentCalls.add(call);

		Log.d(TAG, "create a new Call object " + call + " - " + mCurrentCalls.size());

		// PeerConnection created after CALL_CONFIG message with Trickle ICE, but now with standard call flow.
		
		if (!trickleIceStrategy)
		{
			call.getPeerConnectionClient().createPeerConnection(addVideo, call, null);
		}
		
		return call;
	}

	/**
	 * TODO: We're now supporting multiple calls, so we need to update this method to actually
	 * return the current call and not simply the first in the list.
	 * @return
	 */
	CallImpl getCurrentCallImpl()
	{
		Log.d("CallManager", "getCurrentCallImpl " + mCurrentCalls + " - " + mCurrentCalls.size() + " cm= " + this);

		return ((mCurrentCalls != null) && !mCurrentCalls.isEmpty())
				? (CallImpl)mCurrentCalls.get(0) : null;
	}

	/**
	 *
	 * @param ID
	 * @return
	 */
	private CallImpl getCallImplWithID(String ID)
	{
		Log.d("CallManager", "getCallImplWithID " + mCurrentCalls + " - " + mCurrentCalls.size() + " - " + ID + " cm= " + this);
		
		Log.d(TAG, "getCallImplWithID() - ID argument=" + ID);

		if ( (mCurrentCalls != null) && (ID != null) )
		{
			Call	call = null;
			Iterator<Call> it = mCurrentCalls.iterator();
			while ( it.hasNext() )
			{
				call = it.next();
				if ( call.getCallId().equals(ID) )
					return (CallImpl)call;
			}
		}

		Log.d(TAG, "getCallImplWithID() -Failed to match callIds - returning null...");
		return null;
	}

	/**
	 * @param call
	 */
	public void endCall(Call call)
	{
		Log.d(TAG, "endCall " + call);

		sendServerMessage(CallServerMessageBuilder.createEndMessage(call.getCallId()));

		doEndCall(call);
	}

	public void holdCall(CallImpl call)
	{
		Log.d(TAG, "holdCall " + call);
		
		createAndSendHoldOffer(call);
	}
	
	public void resumeCall(CallImpl call)
	{
		Log.d(TAG, "resumeCall " + call);
		
		sendResume(call);
	}

    public void processIncomingCall(final CallImpl incomingCall)
	{
        //We do not support concurrent calls, and so all calls other than the one being 
        //answered must be ended here.
        //
        //The calls to be ended are ended in a separate loop below, to prevent 
        //ConcurrentModificationException on mCurrentCalls.
        List<Call> callsToEnd = new ArrayList<Call>();
        for (Call nextCall : mCurrentCalls)
        {
            if (!nextCall.getCallId().equals(incomingCall.getCallId()))
            {
                callsToEnd.add(nextCall);
            }
        }
        for (Call callToEnd : callsToEnd) 
        {
            Log.d(TAG, "Ending call " + callToEnd.getCallId() + " to prevent concurrent calls "
                    + "(When answering call " + incomingCall.getCallId() + ")");
            callToEnd.end();
        }
        
		// if we cannot get the camera answer the call without video
		if (!cameraAnalyser.isCameraAvailable(mCamera))
		{
			Log.i(TAG, "Received call but camera is not available");
			incomingCall.setHasLocalVideo(false);
		}

        if (incomingCall.getSDPState() == SDPState.OFFERING)
        {
            //this is an incoming call that started with an offer request, so need
            //to generate the offer now.
            incomingCall.getPeerConnectionClient().createOffer(new SdpObserver() {
                @Override public void onSetSuccess() {}
                @Override public void onSetFailure(String arg0) {}
                @Override public void onCreateSuccess(SessionDescription arg0) {
                    processCall(incomingCall);
                }
                @Override public void onCreateFailure(String desc) {
                    Log.w(TAG, "Failed to create offer for incoming call: " + desc);
                }
            });
        }
        else
        {
            processCall(incomingCall);
        }
	}
	
	/**
	 * @param call
	 */
	public void processCall(final CallImpl call)
	{
		if (trickleIceStrategy)
		{
			// Send OFFER without candidates i.e. trickle ICE.
			considerState(call);
		}
		else
		{
			if (call.getPeerConnectionClient().candidatesComplete )
				considerState(call);
			else
			{
				Log.w(TAG, "Ice candidate gathering is not complete yet, so we can't progress the " +
						"call yet");
				mIceCandidateTimer = new Timer();
				mIceCandidateTimer.schedule(new TimerTask()
				{
					@Override
					public void run()
					{
						if ( call.getPeerConnectionClient().candidatesComplete )
						{
							Log.i(TAG, "IceCandidateChecker::run - candidatesComplete");
	
							considerState(call);
	
							mIceCandidateTimer.cancel();
							mIceCandidateTimer.purge();
							mIceCandidateTimer = null;
						}
					}
				}, 250, 500);
			}		
		}
	}

	/**
	 * @param call
	 */
	protected void considerState(CallImpl call)
	{
		Log.d(TAG, "considerState");

		if ( (call.getPeerConnectionClient() == null) || !call.getPeerConnectionClient().candidatesComplete )
		{
			Log.v(TAG, "Waiting on candidates...");
			return;
		}

        if ( mLocalMediaStream == null )
        {
            Log.v(TAG, "Waiting on media stream");
            return;
        }
        
        if ( (call.getCallStatus() == CallStatus.SETUP) || (call.getSDPState() == SDPState.OFFERING) )
        {
            sendOffer(call);
        }
        else 
        {
            sendAnswer(call);
        }
	}

	/**
	 *
	 */
	private void handleGlare()
	{
		Log.d(TAG, "handleGlare");

		// We've received an offer or offer request so need to roll back the offer
		mOutboundOfferPending = false;
	}
	
	/**
	 * Send a pong on receipt of a ping
	 */
	public void handlePing()
	{
	    Log.d(TAG, "Received a PING...sending PONG");
	    sendServerMessage(CallServerMessageBuilder.createPongMessage());
	}

	/**
	 * @param callID
	 * @param remoteAddress
	 * @param sdp
	 */
	public void handleOffer(final String callID, String remoteParty, final String sdp)
	{
		Log.d(TAG, "handleOffer: callID= " + callID + " remoteParty=" + remoteParty + " sdp=" + sdp);

		handleOfferWithPostRingingCallback(callID, remoteParty, sdp, new SetupHandler()
		{
			@Override
			public void handleSetup(final CallImpl call)
			{
				Log.v(TAG, "handleOffer: handleSetup " + call);
				if ( call != null )
				{
					sendRinging(call);
					
					call.getPeerConnectionClient()
						.setRemoteDescription(CallServerMessageBuilder.MESSAGE_TYPE_OFFER, sdp);
					call.setStatus(CallStatus.ALERTING);
					call.onAlerting();
				}
			}
		});

		if ( mQueuedRemoteNameChange != null )
		{		
		    Log.d(TAG, "About to set display name of caller on call. RemoteDisplayName: " + mQueuedRemoteNameChange);
		    
			final Timer	displayNameTimer = new Timer("handleDisplayNameTimer");
			displayNameTimer.schedule(new TimerTask()
			{
				@Override
				public void run()
				{
					final CallImpl call = getCallImplWithID(callID);
					if (call == null)
					{
						// Try a little bit later, hopefully after the OFFER is received and the call is created.
						Log.d(TAG, "Call not created yet, waiting little while before trying again...");
					}
					else
					{
					    doHandleDisplayName(call, mQueuedRemoteNameChange);
					    mQueuedRemoteNameChange = null;
						Log.d(TAG, "Display name has been set");
					    // exit task
					    this.cancel();
					}
				}
			}, 0, 500); // delay for 0ms, then keep trying every 500ms until call is created			
		}
		else
		{
		    Log.d(TAG, "No display name to set on call");
		}
	}
	
	/**
	 *
	 * @param jsonResult
	 */
    private void handleUpdatedOffer(final CallImpl call, String sdp)
    {
		Log.d(TAG, "handleUpdatedOffer");
        Log.d(TAG, "Offer during call.  SDP was: " + sdp);

        call.getPeerConnectionClient().setRemoteDescription(CallServerMessageBuilder.MESSAGE_TYPE_OFFER, sdp);
        
        final SdpObserver observer = new SdpObserver() {
            
            @Override
            public void onSetSuccess() {
            }
            
            @Override
            public void onSetFailure(String arg0) {
            }
            
            @Override
            public void onCreateSuccess(SessionDescription sdp) 
            {
                sendAnswer(call);      
            }
            
            @Override
            public void onCreateFailure(String arg0) {
            }
        }; 
        
        //If the local side is holding, we should answer with hold SDP
        //if the remote side sends a non hold SDP offer (so a call is not
        //resumed when one side resumes in a 2 way hold scenario
        if ((call.isLocalHeld())
                && !isHoldOffer(sdp))       
        {
            call.getPeerConnectionClient().createForcedHoldAnswer(observer);
        }
        else
        {
            call.getPeerConnectionClient().createAnswer(observer);
        }
    }

	/**
	 * @param callId
	 * @param remoteAddress
	 */
	public void handleOfferRequest(String callId, String remoteAddress)
	{
        Log.d(TAG, "handleOfferRequest");
        
        final CallImpl call = getOrCreateCall(callId, true);
        call.setRemoteAddress(remoteAddress);
        
        call.setSDPState(SDPState.OFFERING);
        
        if (call.getCallStatus().ordinal() > CallStatus.SETUP.ordinal())
        {
            //in-call offer request
            if (call.isLocalHeld())
            {
                createAndSendHoldOffer(call);
            }
            else
            {
                createAndSendOffer(call);
            }
        }
        else
        {
            //this is a new incoming call
            call.onAlerting();
            sendRinging(call);
        }
	}

    private void handleOfferWithPostRingingCallback(final String callId, final String remoteAddress,
        	final String sdp, final SetupHandler handler)
    {
    	trickleIceStrategy = adoptTrickleIceStrategy();

    	Log.d(TAG, "handleOfferWithPostRingingCallback " + trickleIceStrategy);
    	
    	if (trickleIceStrategy)
    	{
    		handleOfferWithPostRingingCallbackTrickleIce(callId, remoteAddress, sdp, handler);
    	}
    	else
    	{
    		handleOfferWithPostRingingCallbackStandard(callId, remoteAddress, sdp, handler);
    	}
    }
    
    		
    private void handleOfferWithPostRingingCallbackStandard(final String callId, final String remoteAddress,
    	final String sdp, final SetupHandler handler)
    {
		Log.d(TAG, "handleOfferWithPostRingingCallbackStandard mOutboundOfferPending=" + mOutboundOfferPending);

//        SessionDescription	description =
//        		new SessionDescription(SessionDescription.Type.OFFER, sdp);
        boolean	hasVideo = true;//description.containsVideo();

        Log.v(TAG, "OFFER " + (hasVideo ? "contains" : "does not contain") + " video");

		final CallImpl	call = getOrCreateCall(callId, hasVideo);

		switch ( call.getCallStatus() )
		{
			case UNINITIALIZED:
				Log.v(TAG, "Current(new) call state is UNINITIALIZED " + call);
				call.setRemoteAddress(remoteAddress);
				handler.handleSetup(call);
				break;

			case IN_CALL:
			case MEDIA_PENDING:
				if ( mOutboundOfferPending )
		            handleGlare();
				else
				{
					handleUpdatedOffer(call, sdp);
				}
				break;
				
			default:
				Log.w(TAG, "Received OFFER in wrong state");
				break;
		}
		
    }

    private void handleOfferWithPostRingingCallbackTrickleIce(final String callId, final String remoteAddress,
    	final String sdp, final SetupHandler handler)
    {
		Log.d(TAG, "handleOfferWithPostRingingCallbackTrickleIce mOutboundOfferPending=" + mOutboundOfferPending);

//        SessionDescription	description =
//        		new SessionDescription(SessionDescription.Type.OFFER, sdp);
        final boolean	hasVideo = true;//description.containsVideo();

        Log.v(TAG, "OFFER " + (hasVideo ? "contains" : "does not contain") + " video");

        // Wait for CALL_CONFIG before creating the call and PeerConnection.
        
		new Thread(new Runnable()
		{
			@Override
			public void run()
			{
			    int waitCount = 0;
				while (iceServers.get(callId) == null)
				{
				    if (waitCount == CALL_CONFIG_RETRIES_PER_SECOND * 10)
				    {
				        // If no call config after 10 seconds give up
				        Log.w(TAG, "No call configuration recieved after 10 seconds... ");
				        return;
				    }
					
				    if (waitCount % CALL_CONFIG_RETRIES_PER_SECOND == 0)
					{   
					    // Log every second
					    Log.d(TAG, "Waiting for call configuration...");
					}
					
	                waitCount++;
					try 
					{
						Thread.sleep(SECOND_IN_MS/CALL_CONFIG_RETRIES_PER_SECOND);
					} 
					catch (InterruptedException e) {}
				}
				
				final CallImpl	call = getOrCreateCall(callId, hasVideo);

				switch ( call.getCallStatus() )
				{
					case UNINITIALIZED:
						Log.v(TAG, "Current(new) call state is UNINITIALIZED " + call);
						call.setRemoteAddress(remoteAddress);
						handler.handleSetup(call);
						break;

					case IN_CALL:
						Log.v(TAG, "Current call state is IN_CALL");
						if ( mOutboundOfferPending )
				            handleGlare();
						else
						{
							handleUpdatedOffer(call, sdp);
						}
						break;

					case MEDIA_PENDING:
						Log.i(TAG, "Current call state is MEDIA_PENDING and we have received an OFFER");
						
						// Wait until we are IN_CALL before attempting to handle the OFFER.
						
						new Thread(new Runnable()
						{
							@Override
							public void run()
							{				
								CallStatus status;
								
								int count = 0;
								while ((status = call.getCallStatus()).ordinal() < CallStatus.IN_CALL.ordinal())
								{
									try 
									{
										Thread.sleep(250);
									} 
									catch (InterruptedException e) {}
								
									// Only log every second we are in this state
									if (count % 4 == 0)
									{
									    Log.d(TAG, "Waiting for call status of IN_CALL before handling OFFER (currently " + status + ")");
									}
									count++;
								}
																
                                if (call.getCallStatus() == CallStatus.IN_CALL)
                                {
                                    Log.i(TAG, "Call status now IN_CALL therefore handling OFFER");
                                    
                                    handleOfferWithPostRingingCallbackStandard(callId, remoteAddress, sdp, handler);
                                }
                                else
                                {
                                	Log.w(TAG, "Call status now " + status + "therefore NOT handling OFFER");
                                }
							}
						}).start();
						
						break;
						
					default:
						Log.w(TAG, "Received OFFER in wrong state");
						break;
				}
				
			}
		}).start();
           
    }
    
    private boolean isHoldOffer(String sdp)
    {
    	if (!sdp.contains("sendrecv")
    			&& !sdp.contains("recvonly"))
    	{
    		return true;
    	}
    	
    	return false;
    }

	/**
	 * An ANSWER message has been received from the server.
	 * @param callID
	 * @param sdp
	 */
	public void handleAnswer(String callID, String sdp)
	{
		Log.d(TAG, "handleAnswer");

		CallImpl	call = getCallImplWithID(callID);
		if ( call != null )
		{
			if ( call.getSDPState() == SDPState.OFFERED )
			{
				call.getPeerConnectionClient().setRemoteDescription(CallServerMessageBuilder.MESSAGE_TYPE_ANSWER,
				                                     sdp);

				call.setSDPState(SDPState.ANSWERED);
				call.updateStatusAndCallback();

				sendOK(call);
				mOutboundOfferPending = false;
			}
			else
			{
				if ( !call.mIsCaller )
					sendOK(call);
				else
					call.onError(CallServerMessageBuilder.MESSAGE_ERROR_TYPE_FAILED,
					             "Received ANSWER in wrong state", null);
			}
		}
	}

	/**
	 * @param callId The id of the call that has errored
	 * @param errorType Error identifier
	 * @param reason Reason error occurred
	 */
	public void handleError(String callId, String errorType, String reason)
	{
		Log.d(TAG, "handleError");

		Call	call = getCallImplWithID(callId);
		if ( call == null )
		{
			if ( CallServerMessageBuilder.MESSAGE_ERROR_TYPE_BUSY.equals(errorType)
					&& "Call Terminated by Callee".equals(reason) )
			{
				/*
				 * We get this error when the user rejects an incoming call, so technically
				 * it's not an error.
				 */
				Log.i(TAG, "Call rejected");
			}
			else
				Log.e(TAG, "[Current call is null] The reason is, '" + reason + "'");
		}
		else
		{
			handleErrorsInCall(call, errorType, reason);
			endCall(call);
		}
	}

	/**
     * @param callId The call that has errored
	 * @param errorType Error identifier
	 * @param reason Reason error occurred
	 */
    private void handleErrorsInCall(Call call, String errorType, String reason)
    {
    	CallImpl	callImpl = (CallImpl)call;
        if ( CallServerMessageBuilder.MESSAGE_ERROR_TYPE_FAILED.equals(errorType) )
        {
            generalCallError(callImpl, CallServerMessageBuilder.MESSAGE_ERROR_TYPE_FAILED,
                             "Failed error message retrieved.", reason);
        }
        else if ( CallServerMessageBuilder.MESSAGE_ERROR_TYPE_BUSY.equals(errorType) )
        {
        	callImpl.setStatus(CallStatus.ERROR);
        	clearErrorTimer();
        	callImpl.onBusy();
        }
        else if ( CallServerMessageBuilder.MESSAGE_ERROR_TYPE_TIMEOUT.equals(errorType) )
        {
        	callImpl.setStatus(CallStatus.ERROR);
        	clearErrorTimer();
        	callImpl.onTimeout();
        }
        else if ( CallServerMessageBuilder.MESSAGE_ERROR_TYPE_NOT_FOUND.equals(errorType) )
        {
        	callImpl.setStatus(CallStatus.ERROR);
        	clearErrorTimer();
        	callImpl.onNotFound();
        }
        else if ( CallServerMessageBuilder.MESSAGE_ERROR_TYPE_NO_MATCH.equals(errorType) )
        {
            generalCallError(callImpl, CallServerMessageBuilder.MESSAGE_ERROR_TYPE_NO_MATCH,
                             "Call failed",
                             "Invalid offer received by the server: " + errorType + " " + reason);
        }
        else if(PeerConnectionClient.MESSAGE_ERROR_MEDIA_UNAVAILABLE.equals(errorType)) {
            callImpl.setStatus(CallStatus.MEDIA_UNAVAILABLE);
            clearErrorTimer();
            callImpl.onMediaUnavailable();
        }
        else
        {
            generalCallError(callImpl, CallServerMessageBuilder.MESSAGE_ERROR_TYPE_FAILED,
                             "The call failed with an unknown error type.", reason);
        }
    };

	/**
	 * @param callId
	 */
	public void handleRinging(String callId)
	{
		Log.d(TAG, "handleRinging");

		CallImpl	call = getCallImplWithID(callId);
		if ( (call != null) && call.isDialling() )
		{
		 	call.setConnectedness(CallImpl.Connectedness.RINGING);
		 	call.updateStatusAndCallback();
		}
	}

	/**
	 *
	 * @param callId
	 */
	public void handleEstablished(String callId)
	{
		Log.d(TAG, "handleEstablished");

		CallImpl	call = getCallImplWithID(callId);
        if ( call != null )
        {
            call.setConnectedness(CallImpl.Connectedness.CONNECTED);
            call.updateStatusAndCallback();
        }
	}

	/**
	 * @param callID
	 *
	 */
	public void handleEnd(String callID)
	{
		Log.d(TAG, "handleEnd");

		Call	call = getCallImplWithID(callID);
		if ( call != null )
		{
			((CallImpl)call).setStatus(CallStatus.ENDED);
			doEndCall(call);
		}
	}

	/**
	 *
	 * @param call
	 */
	private void doEndCall(Call call)
	{
		Assert.assertNotNull(call);

		((CallImpl)call).getPeerConnectionClient().closePeerConnection();

		mCurrentCalls.remove(call);
		
	      // will receive a new CallConfig for each call.
        iceServers.remove(call.getCallId());
        
		call = null;

		Log.i(TAG, "call ended");
	}

	/**
	 * @param callId
	 * @param video
	 * @return
	 */
	private CallImpl getOrCreateCall(String callId, final boolean video)
	{
		CallImpl call = getCallImplWithID(callId);
		
		if ( call == null )
		{
			Log.w(TAG, "No current call exists, so let's create one now.");
			call = createNewCall(null, video); // TODO: need to know here if we need a VideoView
			call.setID(callId); // give the new call the ID that we were given by the server
			
			// This method called for an incoming call.
			
			// This method is called after CALL_CONFIG message with Trickle ICE. 
			// This method is called earlier with standard call flow.
			
			if (trickleIceStrategy)
			{
				call.getPeerConnectionClient().createPeerConnection(video, call, iceServers.get(call.getCallId()));
			}
		}
		
		return call;
	}

	/**
	 *
	 * @param call
	 * @param remoteName
	 */
	protected void doHandleDisplayName(CallImpl call, String remoteName)
	{
		Assert.assertNotNull(call);
		call.setRemoteDisplayName(remoteName);
		call.onRemoteDisplayNameChanged();
	}

	/**
	 * @param callId
	 * @param name
	 */
	public void handleDisplayName(String callId, String name)
	{
		Log.d(TAG, "handleDisplayName");

		/*
		 * Give the new remote display name to the given call, but if there isn't a current
		 * call, we assume that we're going to receive an OFFER imminently, so wait a while and
		 * try again.
		 */
		CallImpl	call = getCallImplWithID(callId);
		if ( call != null )
			doHandleDisplayName(call, name);
		else
		{
			Log.w(TAG, "We don't have a call object yet, so schedule this name update for later");
			mQueuedRemoteNameChange = name;
//			/*
//			 * Schedule it for a little bit later, hopefully after the OFFER is received and the
//			 * call is created.
//			 */
//			Timer	displayNameTimer = new Timer("handleDisplayNameTimer");
//			displayNameTimer.schedule(new TimerTask()
//			{
//				@Override
//				public void run()
//				{
//					doHandleDisplayName(getCurrentCall(), remoteName);
//				}
//			}, 100);
		}
	}
	
	public void handleRemoteHeld(String callID)
    {
	    CallImpl    call = getCallImplWithID(callID);
        if ( call != null )
        {
            call.onRemoteHeld();
        }
    }
    
    public void handleRemoteUnheld(String callID)
    {
        CallImpl    call = getCallImplWithID(callID);
        if ( call != null )
        {
            call.onRemoteUnheld();
        }
    }
	
	private boolean sendResume(final CallImpl call)
	{
		Log.d(TAG, "sendResume");
		
		return createAndSendOffer(call);
	}

	private boolean createAndSendOffer(final CallImpl call)
	{
		Log.d(TAG, "createAndSendOffer ");

	    call.getPeerConnectionClient().createOffer(new SdpObserver() {
            
            @Override
            public void onSetSuccess() {
                // TODO Auto-generated method stub
                
            }
            
            @Override
            public void onSetFailure(String arg0) {
                // TODO Auto-generated method stub
                
            }
            
            @Override
            public void onCreateSuccess(SessionDescription arg0) 
            {
                sendOffer(call);
                
            }
            
            @Override
            public void onCreateFailure(String arg0) {
                // TODO Auto-generated method stub
                
            }
        });
	    
	    return true;
	}
	
	private boolean createAndSendHoldOffer(final CallImpl call)
	{
	    call.getPeerConnectionClient().createHoldOffer(new SdpObserver() {
            
            @Override
            public void onSetSuccess() {
                // TODO Auto-generated method stub
                
            }
            
            @Override
            public void onSetFailure(String arg0) {
                // TODO Auto-generated method stub
                
            }
            
            @Override
            public void onCreateSuccess(SessionDescription arg0) 
            {
                sendOffer(call);
            }
            
            @Override
            public void onCreateFailure(String arg0) {
                // TODO Auto-generated method stub
                
            }
        });
	    
	    return true;
	}
	
	/**
	 *
	 * @param call the call
	 * @param sdp the offer SDP
	 */
	private void sendOffer(final CallImpl call)
	{
		Log.d(TAG, "sendOffer");

    	final String	ipAddress = getLocalIpAddress();
    	if ( TextUtils.isEmpty(ipAddress) )
    	{
    		return;
    	}
    	
    	final String sdp =  getSdpToSend(call);
        sendServerMessage(CallServerMessageBuilder.createOfferMessage(call.getCallId(),
        		sdp));
        
        Log.d(TAG, "SDP = " + sdp);

        call.setSDPState(CallImpl.SDPState.OFFERED);
		if ( call.getCallStatus() == CallStatus.SETUP )
			call.updateStatusAndCallback();
		
		mOutboundOfferPending = true;
	}
	
	/**
	 * Generate the SDP that is to be sent
	 * 
	 * @param call The call
	 * @return
	 */
	private String getSdpToSend(CallImpl call)
	{
		if (call.hasLocalVideo())
		{
		    return addImageAttrToSdp(
	                call.getPeerConnectionClient().getLocalDescription());
		}		
		return call.getPeerConnectionClient().getLocalDescription();
	}
	
	/**
	 * Add imageattr to the SDP.  Ideally, this would be done from within the PeerConnectionClient after SDP
	 * was generated, but it seems to strip the imageattr value after setting the local description.
	 * 
	 * @param sdp The original sdp
	 * @return SDP with imageattr
	 */
	private String addImageAttrToSdp(String sdp)
	{
	    final String recvSize = "x=640,y=480";
	    final PhoneVideoCaptureResolution sendResolution = getClosestSupportedResolutionToPreferred();
	    
	    final StringBuilder builder = new StringBuilder("a=imageattr:* send [x=");
	    builder.append(sendResolution.getWidth());
	    builder.append(",y=");
	    builder.append(sendResolution.getHeight());
	    builder.append("] recv [");
	    builder.append(recvSize);
	    builder.append("]\r\n");
	    
	    return sdp + builder.toString();
	}
	
	private void sendAnswer(final CallImpl call, String sdp)
	{
	    Log.d(TAG, "sendAnswer");

        if ( call.getCallStatus() == CallStatus.ENDED )
        {
            Log.w(TAG, "Call already ended");
        }
        else
        {
            sendServerMessage(CallServerMessageBuilder.createAnswerMessage(call.getCallId(),
                                                           sdp));

            call.setSDPState(SDPState.ANSWERED);
            call.updateStatusAndCallback();
        }
	}

	/**
	 *
	 * @param call
	 */
	private void sendAnswer(final CallImpl call)
	{
	    call.getPeerConnectionClient().createAnswer(new SdpObserver(){
            @Override public void onCreateFailure(String arg0) {}
            @Override public void onSetFailure(String arg0) {}
            @Override public void onSetSuccess() {}
        
            @Override
            public void onCreateSuccess(SessionDescription arg0) 
            {
                final String sdp = getSdpToSend(call);
                sendAnswer(call, sdp);
            }            
        });
	}

    /**
     *
     * @param call
     */
    private void sendOK(final CallImpl call)
    {
		Log.d(TAG, "sendOK");

        sendServerMessage(CallServerMessageBuilder.createOkMessage(call.getCallId()));

		if ( call.getCallStatus() == CallStatus.ALERTING )
			call.setStatus(CallStatus.MEDIA_PENDING);
    }

    /**
     *
     * @param call
     */
	protected void sendRinging(final Call call)
    {
		Log.d(TAG, "sendRinging");

    	sendServerMessage(CallServerMessageBuilder.createRingingMessage(call.getCallId()));
    }

	/**
	 * @param call
	 * @param code
	 */
	public void sendDTMFCode(final Call call, String code)
	{
		Log.d(TAG, "sendDTMFCode");

		sendServerMessage(CallServerMessageBuilder.createDtmfMessage(call.getCallId(), code));
	}

	/**
	 *
	 * @return
	 */
	private static String getLocalIpAddress()
	{
	    try
	    {
	        for ( Enumeration<NetworkInterface> interfaces =
	        			NetworkInterface.getNetworkInterfaces();
	        		interfaces.hasMoreElements(); )
	        {
	            NetworkInterface intf = interfaces.nextElement();
	            for ( Enumeration<InetAddress> addresses = intf.getInetAddresses();
	            		addresses.hasMoreElements(); )
	            {
	                InetAddress inetAddress = addresses.nextElement();
	                if ( !inetAddress.isLoopbackAddress() )
	                {
	                    return inetAddress.getHostAddress().toString();
	                }
	            }
	        }
	    }
	    catch (SocketException ex)
	    {
	        Log.e(TAG, ex.toString());
	    }

	    return null;
	}

	/**
	 * 
	 * @param call The call that has errored
	 * @param error Error identifier
	 * @param message Message detailing error
	 * @param reason Reason error occurred
	 */
	private void generalCallError(final CallImpl call, final String error, final String message,
		final String reason)
	{
		Log.v(TAG, "generalCallError. " + error + ", " + message + ", " + reason);

		/*
		 * General errors may come around the same time as a more detailed explanation, so delay
		 * the general error, and don't send if more detail did come.
		 */
		mGeneralErrorTimer = new Timer();
		mGeneralErrorTimer.schedule(new TimerTask()
		{
			@Override
			public void run()
			{
				Log.d(TAG, "[generalCallError] " + message + ":" + reason);

				mGeneralErrorTimer = null;

				if ( call != null )
				{
					// Have we already flagged an error?
					if ( call.getCallStatus() != CallStatus.ERROR )
						call.onError(error, message, reason);
					else
						Log.v(TAG, "Ignoring this error because we're already in the ERROR state");
				}
				else
					Log.w(TAG, "current call is null");
			}
		}, 100);
	}

	/**
	 *
	 */
	private void clearErrorTimer()
	{
		if ( mGeneralErrorTimer != null )
		{
			mGeneralErrorTimer.cancel();
			mGeneralErrorTimer.purge();
			mGeneralErrorTimer = null;
		}
	}

	@Override
	public void addListener(PhoneListener listener)
	{
		if ( (mListeners != null) && (mListeners.contains(listener) == false) )
			mListeners.add(listener);
	}

	@Override
	public void removeListener(PhoneListener listener)
	{
		if ( mListeners != null )
			mListeners.remove(listener);
	}

	@Override
	public List<Call> getCurrentCalls()
	{
		return mCurrentCalls;
	}
	
	/*
	 * Returns current active call, or null if there is not one.
	 */
	private CallImpl getActiveCall()
	{
	    if (mCurrentCalls != null && mCurrentCalls.size() > 0)
	    {
	        // previous code always returned the first one; do the same for now
	        return (CallImpl)mCurrentCalls.get(0);
	    }
	    else
	    {
	        return null;
	    }
	}

	@Override
    public void setPreviewView(VideoSurface view)
    {
		mPreviewView = view;
		LocalMediaManager.getInstance().setLocalVideoView(view);
    }

	@Override
	public void setVideoOrientation(int orientation)
	{
	    mOrientation = orientation;
	    
	    LocalMediaManager.getInstance().setVideoOrientation(orientation);
	}

	@Override
	public void setCamera(int camera)
	{
		Log.d(TAG, "setCamera() called with cameraIdz: " + camera);
		
	    final CallImpl activeCall = getActiveCall();	    
	    if (activeCall != null)
	    {
	        final PeerConnectionClient peerConnection = activeCall.getPeerConnectionClient();
	        if (peerConnection != null)
	        {
	            peerConnection.setCameraIndex(camera);
	        }
	        
	        // WEBRTC-7147 in-call camera switch
	        final VideoCaptureAndroid instance = VideoCaptureAndroid.getInstance();
	        if (instance != null)
	        {
	        	instance.changeCamera(camera);
	        }
	        else
	        {
	        	Log.d(TAG, "Could not find Video capture instance - cannot set Camera");
	        }
	    }
	    
	    // always set this ready for next call
	    mCamera = camera;
	}
	
	/**
     * It is only valid to create a call if there is not already one in progress.
	 * 
     * @return True if a new call can be created, false otherwise.
	 */
    private boolean isValidToCreateNewCall()
	{
        return mCurrentCalls.isEmpty();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.alicecallsbob.fcsdk.android.phone.Phone#createCall(java.lang.String, boolean, boolean, com.alicecallsbob.fcsdk.android.phone.CallListener)
	 */
	@Override
	public Call createCall(final String address, final boolean audio, final boolean video,
		CallListener listener) throws CallCreationWithErrorException
	{
        if (!isValidToCreateNewCall())
        {
            throw new CallCreationWithErrorException("Cannot create new call as there is "
                    + "already an active call.", CallCreationError.CALL_IN_PROGRESS);
        }

		// throw error if the camera is in use by another app, NOT if it is in use on another call
		if (video && !isCameraInUseOnAnotherCall() && !cameraAnalyser.isCameraAvailable(mCamera))
		{
			throw new CallCreationWithErrorException(
			        "Cannot get device camera, it may be in use by another application",
			        CallCreationError.CAMERA_UNAVAILABLE);
		}

		// throw error if the microphone is in use by another app, NOT if it is in use on another call
		if (!isMicrophoneInUseOnAnotherCall() && MicrophoneAnalyser.isMicrophoneInUse())
		{
			throw new CallCreationWithErrorException(
			        "Cannot use Microphone, it may be in use by another application",
			        CallCreationError.MICROPHONE_UNAVAILABLE);
		}

        trickleIceStrategy = adoptTrickleIceStrategy();
		if (trickleIceStrategy)
		{
			Log.i(TAG, "Adopting Trickle ICE call strategy.");
			
			return createCallTrickleIce(address, audio, video, listener);
		}
		else
		{
			// Default to pre-2.1.8 behaviour.
			// This will also be the behaviour if swiftVersion is 0 as expected when talking to older servers.
			Log.i(TAG, "Adopting standard call strategy.");
			
			return createCallStandard(address, audio, video, listener);
		}
	}
	    
	private boolean isMicrophoneInUseOnAnotherCall()
    {
		for (Call call : mCurrentCalls)
		{
			if (call.getCallStatus() == CallStatus.IN_CALL)
			{
				return true;
			}
		}
		return false;
    }
	
	private boolean isCameraInUseOnAnotherCall()
	{
		for (Call call : mCurrentCalls)
		{
			if (call.getCallStatus() == CallStatus.IN_CALL && call.hasLocalVideo())
			{
				return true;
			}
		}
		return false;
	}	

	private boolean adoptTrickleIceStrategy() 
	{
		Log.i(TAG, "adoptTrickleIceStrategy SWIFT version=" + swiftVersion);
		return (swiftVersion >= 2) ? true : false;
	}
	
	public Call createCallStandard(final String address, final boolean audio, final boolean video,
			CallListener listener) throws CallCreationWithErrorException
	{
		Log.d(TAG, "createCall standard");
		
		final CallImpl	call = createNewCall(address, video, listener);
		if ( call == null )
			return null;

		//call.addListener(listener);

		// FIXME: Why is this a separate thread? Could it be effecting the crashes I'm getting?
		new Thread(new Runnable()
		{
			@Override
			public void run()
			{
				if ( call.setup(audio, video) )
				{
					Log.d(TAG, "connectCall");
					call.setConnectedness(CallImpl.Connectedness.DIALLING);
			        sendServerMessage(CallServerMessageBuilder
			                          .createConnectMessage(call.getCallId(), address));
					call.getPeerConnectionClient().createOffer();
					
					// Not Trickle ICE.
					call.getPeerConnectionClient().candidatesComplete = false;
					
				    processCall(call);
				}
			}
		}, "Connecting").start();

		return call;
	}	
	
	public Call createCallTrickleIce(final String address, final boolean audio, final boolean video,
		CallListener listener) throws CallCreationWithErrorException
	{
		Log.d(TAG, "createCall Trickle ICE");
		
		final CallImpl	call = createNewCall(address, video, listener);
		if ( call == null )
			return null;

		Log.d(TAG, "connectCall");
		call.setConnectedness(CallImpl.Connectedness.DIALLING);
        sendServerMessage(CallServerMessageBuilder
                          .createConnectMessage(call.getCallId(), address));
		
		// Wait for CALL_CONFIG before creating PeerConnection.
		new Thread(new Runnable()
		{
			@Override
			public void run()
			{
			    int counter = 0;
				while ((iceServers.get(call.getCallId()) == null) && !call.isEnding() &&
				        call.getCallStatus() != CallStatus.ENDED)
				{
				    // Only log this every second 
				    if (counter % CALL_CONFIG_RETRIES_PER_SECOND == 0)
				    {
				        Log.d(TAG, "Waiting for call configuration...");
				    }
					counter++;
					
					try 
					{
						Thread.sleep(SECOND_IN_MS/CALL_CONFIG_RETRIES_PER_SECOND);
					} 
					catch (InterruptedException e) {}
				}
				
				if (call.isEnding() || call.getCallStatus() == CallStatus.ENDED)
				{
					Log.d(TAG, "Outgoing call has been cancelled whilst waiting for call configuration.");
					return;
				}
				
				final boolean success = call.getPeerConnectionClient().createPeerConnection(video, call, iceServers.get(call.getCallId()));
				
				if (success && !call.isEnding() && call.getCallStatus() != CallStatus.ENDED)
				{
    				if ( call.setup(audio, video) )
    				{
    					call.getPeerConnectionClient().createOffer();
    				    processCall(call);
    				}
				}
			}
		}, "Connecting").start();

		return call;
	}

	/**
     * {@inheritDoc}
     */
    @Override
    public List<PhoneVideoCaptureSetting> getRecommendedCaptureSettings() 
    {
    	final Map<Integer, List<PhoneVideoCaptureSetting>> supportedCameraSettings =
    			cameraAnalyser.getSupportedCameraSettings();
    	
    	if (supportedCameraSettings.size() == 0)
    	{
    		Log.w(TAG, "Could not find any camera capture settings");
    		return new ArrayList<PhoneVideoCaptureSetting>();
    	}
        // if only 1 camera, return first element, else we use the mCamera as the idx into the list
        if (supportedCameraSettings.entrySet().size() == 1)
        {
            return supportedCameraSettings.get(0);            
        }        
        return supportedCameraSettings.get(mCamera);
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public List<PhoneVideoCaptureSetting> getRecommendedCaptureSettings(int camera) 
    {
    	final Map<Integer, List<PhoneVideoCaptureSetting>> supportedCameraSettings =
    			cameraAnalyser.getSupportedCameraSettings();
    	if (supportedCameraSettings.size() <= camera) {
    		Log.w(TAG, "Details requested for unknown camera: " + camera);
    		return new ArrayList<PhoneVideoCaptureSetting>();
    	}

        return supportedCameraSettings.get(mCamera);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setPreferredCaptureResolution(
            PhoneVideoCaptureResolution resolution) 
    {
        mPreferredCaptureResolution = resolution;
        
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setPreferredCaptureFrameRate(int frameRate) 
    {
        mPreferredFramerate = frameRate;
        
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public PhoneVideoCaptureResolution getPreferredCaptureResolution() 
    {
        return mPreferredCaptureResolution;
    }    
        
    @Override
    public int getCamera() 
    {
        return mCamera;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int getPreferredCaptureFrameRate() 
    {
        return mPreferredFramerate;
    }
        
    /**
     * Returns the closest supported resolution to the preferred resolution.
     * 
     * @return The closest supported resolution to the preferred resolution.
     */
    public PhoneVideoCaptureResolution getClosestSupportedResolutionToPreferred()
    {
        final PhoneVideoCaptureResolution preferred = getPreferredCaptureResolution();
        Log.i(TAG, "Preferred  capture resolution - width=" + preferred.getWidth() 
        		+ " height=" + preferred.getHeight());
        
        final long preferredMultiplied = preferred.getWidth() * preferred.getHeight();
        
        PhoneVideoCaptureResolution currentClosest = null;
        long distanceFromPreferred = -1;
        
        for (PhoneVideoCaptureSetting setting : getRecommendedCaptureSettings())
        {
            final int thisWidth = setting.getResolution().getWidth();
            final int thisHeight = setting.getResolution().getHeight();
            if (thisWidth == preferred.getWidth()
                    && thisHeight == preferred.getHeight())
            {
                currentClosest = setting.getResolution();
                break;
            }
            else
            {
                long thisDistanceFromPreferred = Math.abs((thisWidth * thisHeight) - preferredMultiplied);
                
                if (currentClosest == null
                        || distanceFromPreferred > thisDistanceFromPreferred)
                {
                    distanceFromPreferred = thisDistanceFromPreferred;
                    currentClosest = setting.getResolution();
                }
            }
        }                        
        Log.i(TAG, "Closest supported capture resolution: width=" + currentClosest.getWidth() 
        		+ " height=" + currentClosest.getHeight());
        
        return currentClosest;
    }
    
    /**
     * Calculates the closest supported framework to the preferred rate.
     * 
     * @return The best supported framerate
     */
    public Integer getClosestSupportedFramerateToPreferred()
    {
    	final Map<Integer, List<int[]>> supportedCameraFramerates =
    			cameraAnalyser.getSupportedCameraFramerates();
        return cameraAnalyser.getClosestSupportedFramerate(getPreferredCaptureFrameRate(), 
        		supportedCameraFramerates.get(mCamera));
    }
    
    @Override
    public void enableLocalAudio(boolean isAudioEnabled) 
    {
        LocalMediaManager.getInstance().enableLocalAudio(isAudioEnabled);      
    }

    @Override
    public void enableLocalVideo(boolean isVideoEnabled) 
    {
        LocalMediaManager.getInstance().enableLocalVideo(isVideoEnabled);   
    }
    
    /**
    *
    * @param stream
    */
    void setLocalMediaStream(final MediaStream stream)
    {
        if (mLocalMediaStream != stream)
        {
            mLocalMediaStream = stream;
            for (PhoneListener listener : mListeners)
            {
                if ( listener != null )
                    listener.onLocalMediaStream();
            }
        }
    }
    
    MediaStream getLocalMediaStream()
    {
        return mLocalMediaStream;
    }

	public void onIceCandidate(CallImpl call, IceCandidate candidate) 
	{
		String callID = call.getCallId();
		
		Log.d(TAG, "onIceCandidate " + callID + " " + trickleIceStrategy + " " + candidate);

		if (trickleIceStrategy)
		{
			// Need to send an ICE_CANDIDATE SWIFT message (once call is established).
			trickleIceTimer.scheduleAtFixedRate(new SendIceCandidateTimerTask(call, candidate, this), 500, 250);
		}
		else
		{
			// Standard strategy is to do nothing with individual ICE candidates.
		}
	}

	public void handleIceCandidate(String callID, IceCandidate candidate) 
	{
		Log.d(TAG, "handleIceCandidate " + callID + " " + trickleIceStrategy + " " + candidate);
		
		if (trickleIceStrategy)
		{
			CallImpl call = getCallImplWithID(callID);
		
			if (call.getCallStatus().ordinal() > CallStatus.IN_CALL.ordinal())
			{
				Log.d(TAG, "Call has been terminated so no need to set remote ICE candidate.");
				return;
			}
			
			call.getPeerConnectionClient().onRemoteIceCandidate(candidate);
		}
		else
		{
			// Standard strategy is to do nothing with remote ICE candidates.
		}
	}

	/**
	 * Sends ICE_CANDIDATE message if the call has been connected.
	 */
	static class SendIceCandidateTimerTask extends TimerTask
	{		
		private CallImpl call;
		private IceCandidate candidate;
		private PhoneImpl phone;
		private int waitingToSendIceCounter = 0;
		private static final int iceLogFraction = 4;

		SendIceCandidateTimerTask(CallImpl call, IceCandidate candidate, PhoneImpl phone)
		{
			this.call = call;
			this.candidate = candidate;
			this.phone = phone;
		}
		
		public void run()
		{
			CallStatus callStatus = call.getCallStatus();
			
			Log.d(TAG, String.format("SendIceCandidateTimerTask call status %s(%d)", callStatus, callStatus.ordinal()));
			
			if (callStatus.ordinal() > CallStatus.IN_CALL.ordinal())
			{
				Log.d(TAG, "Call has been terminated so no need to send ICE candidate.");
				cancel();
				return;
			}
			
			if (call.isConnected())
			{
				Log.d(TAG, "SendIceCandidateTimerTask sending " + call.getCallId() + " " + candidate);
				
				phone.sendServerMessage(CallServerMessageBuilder.createIceCandidateMessage(call.getCallId(), candidate));
				cancel();
			}
			else
			{
				waitingToSendIceCounter++;
				if(waitingToSendIceCounter > iceLogFraction) {
				    Log.d(TAG, "SendIceCandidateTimerTask waiting for call establishment before sending ICE candidate " + call.getCallId() + " " + candidate);
				    waitingToSendIceCounter = 0;
				}
			}
		}
	}

	public void handleCallConfig(String callID, List<IceServer> servers) 
	{
	    // This information can't be stored on the call object as it might not have been created yet
		this.iceServers.put(callID, servers);

		Log.d(TAG, "handleCallConfig " + callID + " " + servers);
	}

	public void setSwiftVersion(int version) 
	{
		Log.d(TAG, "setSwiftVersion:" + version);
		swiftVersion = version;
	}
	
	@Override
	public AudioDeviceManager getAudioDeviceManager()
	{
	    return mAudioDeviceManager;
	}
	
}
