/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.uc;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.TrustManager;

import com.alicecallsbob.fcsdk.android.ListenerTarget;
import com.alicecallsbob.fcsdk.android.aed.AED;
import com.alicecallsbob.fcsdk.android.phone.Phone;

/**
 * The UC acts as the entry point to the SDK and provides the client application
 * with a single object exposing all of the other functions through its
 * properties (voice/video, presence/IM and AED). An instance of this class
 * should be created with the {@link UCFactory}.
 */
public interface UC extends ListenerTarget<UCListener>
{
    /**
     * Returns the object for voice/video interaction.
     *
     * @return the object for voice/video interaction.
     */
    Phone getPhone();

    /**
     * Returns the object for AED interaction.
     *
     * @return the object for AED interaction.
     */
    AED getAED();

    /**
     * Connects to the session indicated by the session key provided in the
     * {@link UCFactory#createUc(android.content.Context, String, UCListener)} or
     * {@link UCFactory#createUc(android.content.Context, String, java.util.List, UCListener)}
     * methods.
     */
    void startSession();

    /**
     * Disconnects from the session indicated by the session key provided in the
     * {@link UCFactory#createUc(android.content.Context, String, UCListener)} or
     * {@link UCFactory#createUc(android.content.Context, String, java.util.List, UCListener)}
     * methods.
     */
    void stopSession();

    /**
     * Informs the SDK that the network is reachable or not.
     *
     * @param reachable Whether the network is reachable.
     */
    void setNetworkReachable(boolean reachable);

    /** 	 
     * Set the TrustManager used to validate the connection to the server. 	 
     * 	 
     * @param trustManager a custom TrustManager instance, or null to use the default TrustManager. 	 
     */ 	 
    void setTrustManager(TrustManager trustManager);
    
    /**
     * Sets the HostnameVerifier used to validate the connection to the server. 
     *
     * @param trustManager a custom hostnameVerifier instance, or null to use the default HttpsURLConnection HostnameValidator.
     */
    void setHostnameVerifier(HostnameVerifier hostnameVerifier);
    
    /**
     * Sets if we should send cookies when establishing a connection to the server.
     * Requires that a default CookieHandler has been set in the application to capture cookies from explicit
     * HttpConnections to the server, such as from a Login event.
     * @param use True if cookies should be sent to the server, False if none are needed. Default is False
     */
    void setUseCookies(boolean use);
}
