/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.uc;

import java.util.List;

import com.alicecallsbob.fcsdk.android.impl.ServerMessageManager;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageTransport;
import com.alicecallsbob.fcsdk.android.uc.impl.UCImpl;

import android.content.Context;

/**
 * A class that is responsible for creating instances of the {@link UC}. Creating
 * and starting multiple instances of {@link UC} simultaneously is not recommended.
 * A client should stop one UC session before starting another.
 */
public final class UCFactory
{
    /**
     * Private constructor to suppress javadoc generation.
     */
    private UCFactory()
    {
        super();
    }

    /**
     * Creates a UC instance.
     *
     * @param context The Android Context.
     * @param session The session key to be used with the UC.
     * @param listener The UC listener that will be informed of UC events.
     * @return the UC instance.
     */
    public static UC createUc(Context context, String session, UCListener listener)
    {
       return createUc(context, session, null, listener);
    }

    /**
     * Creates a UC instance using the given STUN servers.
     *
     * @param context The Android Context.
     * @param session The session key to be used with the UC.
     * @param stunServers The STUN servers.
     * @param listener The UC listener that will be informed of UC events.
     * @return the UC instance.
     */
    public static UC createUc(Context context, String session, List<String> stunServers,
        UCListener listener)
    {
        final ServerMessageTransport transport = new ServerMessageTransport();
        final ServerMessageManager serverMessageManager = new ServerMessageManager(transport, session);
        transport.setListener(serverMessageManager);
        return new UCImpl(serverMessageManager, context, stunServers, listener);
    }
}
