/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.uc;

/**
 * A listener for the {@link UC} that will be informed of events.
 */
public interface UCListener
{
    /**
     * Invoked when the session connected successfully.
     */
    void onSessionStarted();

    /**
     * Invoked when the session failed to connect successfully.
     */
    void onSessionNotStarted();

    /**
     * Invoked when the server has experienced a system failure. The application
     * attempt to log in again and re-establish a new session, or direct the
     * user to do so.
     */
    void onSystemFailure();

    /**
     * Invoked when there are problems with the network connection and the
     * connection to the server has been lost. The application should log in
     * again and re-establish a new session, or direct the user to do so.
     */
    void onConnectionLost();
    
    /**
     * Invoked when a connection has been lost and a retry attempt failed.
     * 
     * @param attemptNumber The number of times that the connection has been retried
     * @param delayUntilNextAttempt The time until the next retry attempt
     */
    void onConnectionRetry(int attemptNumber, long delayUntilNextAttempt);
    
    /**
     * Invoked when a failed connection has been reestablished.
     * 
     */
    void onConnectionReestablished();

    /**
     * Invoked when a generic error has occurred when processing a request.
     * @param error
     * @param reason
     */
	void onGenericError(final String error, final String reason);
}
