/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.uc.impl;

import java.util.List;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.TrustManager;

import org.json.JSONObject;

import android.content.Context;
import android.util.Log;

import com.alicecallsbob.fcsdk.android.aed.AED;
import com.alicecallsbob.fcsdk.android.aed.impl.AEDManager;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageListenerTargetImpl;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageManager;
import com.alicecallsbob.fcsdk.android.impl.TransportFailureHandler;
import com.alicecallsbob.fcsdk.android.phone.Call;
import com.alicecallsbob.fcsdk.android.phone.CallStatus;
import com.alicecallsbob.fcsdk.android.phone.Phone;
import com.alicecallsbob.fcsdk.android.phone.impl.CallImpl;
import com.alicecallsbob.fcsdk.android.phone.impl.PhoneImpl;
import com.alicecallsbob.fcsdk.android.uc.UC;
import com.alicecallsbob.fcsdk.android.uc.UCListener;
import com.alicecallsbob.fcsdk.android.uc.impl.handler.GenericErrorHandler;
import com.alicecallsbob.fcsdk.android.uc.impl.handler.InitialisationFailureHandler;
import com.alicecallsbob.fcsdk.android.uc.impl.handler.InitialisationSuccessHandler;
import com.alicecallsbob.fcsdk.android.uc.impl.handler.SystemFailureHandler;

public class UCImpl extends ServerMessageListenerTargetImpl<UCListener> implements UC,
		TransportFailureHandler
{
	private static final String TAG = "UCImpl";

	private PhoneImpl mCallManager;
	private AEDManager mAedManager;
	private final ServerMessageManager mServerMessageManager;
	private final Context mApplicationContext;

	private int swiftVersion = 0;
	
	private final Object lock = new Object();

	
	public UCImpl(final ServerMessageManager serverMessageManager, final Context context,
			final List<String> stunServers, final UCListener listener)
	{
		super(serverMessageManager);
		Log.d(TAG, "Constructor");
		addListener(listener);
		setupHandlers();
		this.mServerMessageManager = serverMessageManager;
		this.mApplicationContext = context;
	}

	private void setupHandlers()
	{
		final InitialisationSuccessHandler successHandler =
				new InitialisationSuccessHandler(this);
		final InitialisationFailureHandler failureHandler =
				new InitialisationFailureHandler(this);
		final SystemFailureHandler systemFailureHandler = new SystemFailureHandler(this);
		final GenericErrorHandler genericErrorHandler = new GenericErrorHandler(this);
		addHandler(UCServerMessageBuilder.MESSAGE_TYPE_INITIALISATION_SUCCESS,
				   successHandler);
		addHandler(UCServerMessageBuilder.MESSAGE_TYPE_INITIALISATION_ERROR,
				   failureHandler);
		addHandler(UCServerMessageBuilder.MESSAGE_TYPE_SYSTEM_FAILURE,
				   systemFailureHandler);
		addHandler(UCServerMessageBuilder.MESSAGE_TYPE_GENERIC_ERROR, genericErrorHandler);
	}

	@Override
	public Phone getPhone()
	{
		synchronized(lock )
		{
			if (mCallManager == null)
			{
				mCallManager = new PhoneImpl(mApplicationContext, mServerMessageManager);
			}
		}
		if (swiftVersion != 0)
		{
			// swift may have already been received
			// due to startSession() being called before this
			mCallManager.setSwiftVersion(swiftVersion);
		}
		
		return mCallManager;
	}

	@Override
	public AED getAED()
	{
		if (mAedManager == null)
		{
			mAedManager = new AEDManager(mServerMessageManager);
		}
		return mAedManager;
	}

	@Override
	public void startSession()
	{
		final JSONObject json = UCServerMessageBuilder.createInitialisationMessage();
		getServerMessageManager().startWithServerMessage(this, json);
	}

	@Override
	public void stopSession()
	{
	    Log.d(TAG, "stopSession");  	        
		getServerMessageManager().stop();

		if (mCallManager != null)
		{
			mCallManager.destroy();
		}

		mAedManager = null;
		mCallManager = null;
	}

	@Override
	public void setNetworkReachable(final boolean reachable)
	{
	    if (!reachable)
	    {
    	    // Cancel all calls that haven't been established as websocket messages
    	    // wont be sent while reachability is set to false
            if (mCallManager != null)
            {
                List<Call> calls = mCallManager.getCurrentCalls();
                for (Call thisCall : calls)
                {
                    if (thisCall.getCallStatus() == CallStatus.UNINITIALIZED
                            || thisCall.getCallStatus() == CallStatus.SETUP
                            || thisCall.getCallStatus() == CallStatus.ALERTING
                            || thisCall.getCallStatus() == CallStatus.RINGING
                            || thisCall.getCallStatus() == CallStatus.MEDIA_PENDING)
                    {
                        ((CallImpl) thisCall).endOnDisconnect();
                    }
                }
            }
	    }
	    
		getServerMessageManager().setNetworkReachable(reachable);
	}
	
	@Override
    public void setTrustManager(TrustManager trustManager) 	 
    { 	 
        getServerMessageManager().setTrustManager(trustManager); 	 
    }
	
	@Override
    public void setHostnameVerifier(HostnameVerifier hostnameVerifier)
    {
		getServerMessageManager().setHostnameVerifier(hostnameVerifier);
    }
	
	@Override
	public void handleTransportFailure()
	{
		Log.v(TAG, "handleTransportFailure");
		
	    if (mCallManager != null)
        {
            List<Call> calls = mCallManager.getCurrentCalls();
            for (Call thisCall : calls)
            {
                ((CallImpl) thisCall).endOnDisconnect();
            }
        }
	      
		for (UCListener listener : listeners())
		{
			listener.onConnectionLost();
		}
	}

	public void onSessionStarted()
	{
		for (UCListener listener : listeners())
		{
			listener.onSessionStarted();
		}

		final JSONObject capabilitiesMsg = UCServerMessageBuilder.createCapabilitiesMessage();
		getServerMessageManager().sendServerMessage(capabilitiesMsg);
	}

	public void onSessionNotStarted()
	{
		for (UCListener listener : listeners())
		{
			listener.onSessionNotStarted();
		}
	}

	public void onSystemFailure()
	{
		for (UCListener listener : listeners())
		{
			listener.onSystemFailure();
		}
	}

	public void onGenericError(final String error, final String reason)
	{
		for (UCListener listener : listeners())
		{
			listener.onGenericError(error, reason);
		}
	}

    @Override
    public void handleTransportRetry(int attemptNumber,
            long delayUntilNextAttempt) 
    {
        for (UCListener listener : listeners())
        {
            listener.onConnectionRetry(attemptNumber, delayUntilNextAttempt);
        }
        
    }

    @Override
    public void handleTransportReestablished() 
    {
        for (UCListener listener : listeners())
        {
            listener.onConnectionReestablished();
        }        
    }
    
    @Override
    public void setUseCookies(final boolean use)
    {
        mServerMessageManager.setUseCookies(use);
    }

	public void setSwiftVersion(int version) 
	{
		// cache the value so we can set it on 
		// any Phones created after this is called
		swiftVersion = version;
		if (mCallManager != null)
		{
			mCallManager.setSwiftVersion(version);
		}	
	}
}
