/*
 * Mon Jun 26 09:41:27 BST 2017 CafeX Communications
 */
package com.alicecallsbob.fcsdk.android.uc.impl;

import java.util.Arrays;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.alicecallsbob.fcsdk.android.SdkVersion;
import com.alicecallsbob.fcsdk.android.impl.ServerMessageBuilder;

import android.os.Build;

public abstract class UCServerMessageBuilder extends ServerMessageBuilder
{
    public static final String MESSAGE_TYPE_INITIALISE				= "INITIALISE";
    public static final String MESSAGE_TYPE_INITIALISATION_SUCCESS	= "INITIALISATION_SUCCESS";
    public static final String MESSAGE_TYPE_INITIALISATION_ERROR	= "INITIALISATION_ERROR";
    public static final String MESSAGE_TYPE_ENDPOINT_CAPABILITIES	= "CAPABILITIES";
    public static final String MESSAGE_TYPE_SYSTEM_FAILURE			= "SYSTEM_FAILURE";
    public static final String MESSAGE_TYPE_GENERIC_ERROR			= "GENERIC_ERROR";

    public static final String	MESSAGE_FIELD_ERROR		= "errorType";
    public static final String	MESSAGE_FIELD_REASON	= "reasonPhrase";

    public static final String	MESSAGE_FIELD_CAPABILITIES								= "capabilities";
    public static final String	MESSAGE_FIELD_CAPABILITIES_REQUIRES_TRICKLE_ICE			= "requires-trickle-ice";
    public static final String	MESSAGE_FIELD_CAPABILITIES_SUPPORTS_BUNDLE				= "supports-bundle";
    public static final String	MESSAGE_FIELD_CAPABILITIES_SUPPORTS_RENEGOTIATION		= "supports-renegotiation";
    public static final String	MESSAGE_FIELD_CAPABILITIES_AVAILABLE_MEDIA_TRANSPORTS	= "available-media-transports";
    public static final String	MESSAGE_FIELD_CAPABILITIES_CLIENT_NAME					= "client-name";
    public static final String	MESSAGE_FIELD_CAPABILITIES_CLIENT_VERSION				= "client-version";
    public static final String	MESSAGE_FIELD_CAPABILITIES_USER_AGENT					= "user-agent";
    
    private static final String USER_AGENT_STRING = "Android " + getOSVersion() + " " + getDeviceName();

    private static final String SDK_VERSION_STRING = getSdkVersionString();
    
    public static JSONObject createInitialisationMessage()
    {
        return createMessageWithType(MESSAGE_TYPE_INITIALISE);
    }

    public static JSONObject createCapabilitiesMessage()
    {
    	try
    	{
    		JSONObject capabilityMap = new JSONObject();
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_REQUIRES_TRICKLE_ICE, false);
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_SUPPORTS_BUNDLE, true);
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_SUPPORTS_RENEGOTIATION, true);
    		final JSONArray mediaTransportArray = new JSONArray(Arrays.asList("WEBRTC_UDP"));
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_AVAILABLE_MEDIA_TRANSPORTS, mediaTransportArray);
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_CLIENT_NAME, "Android");
    		//setting version to zero rather than the real version because the server currently parses
    		//this as an int, not a dotted version string. Change to real version when server supports it.
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_CLIENT_VERSION, SDK_VERSION_STRING);
    		capabilityMap.put(MESSAGE_FIELD_CAPABILITIES_USER_AGENT, USER_AGENT_STRING);

			return createMessageWithType(
					MESSAGE_TYPE_ENDPOINT_CAPABILITIES,
					MESSAGE_FIELD_CAPABILITIES_USER_AGENT, USER_AGENT_STRING,
					MESSAGE_FIELD_CAPABILITIES, capabilityMap);
		}
    	catch (JSONException e)
    	{
			throw new IllegalArgumentException("Failed to build " + MESSAGE_TYPE_ENDPOINT_CAPABILITIES + " SWIFT message.", e);
		}
    }
    
    private static String getOSVersion() 
    {
        return Build.VERSION.RELEASE;
    }
    
    private static String getDeviceName() 
    {
        String manufacturer = Build.MANUFACTURER;
        String model = Build.MODEL;
        if (model.startsWith(manufacturer)) {
            return model;
        } else {
            return manufacturer + " " + model;
        }
    }
    
    /**
     * The server expects the client version to be an int, but the SDK uses a four-part dotted 
     * notation. This method converts the dotted notation to an int, eg, "3.1.2.0" would
     * become "03010200".
     * 
     * @return  The SDK version in the format described above
     */
    private static String getSdkVersionString() 
    {
        final String dotted = SdkVersion.SDK_VERSION_NUMBER;
        final String[] dottedParts = dotted.split("\\.");
        String result = "";
        for (String dottedPart : dottedParts)
        {
            result += (dottedPart.length() == 1) ? ("0" + dottedPart) : dottedPart;
        }
        
        //pad right with '0'. This covers SDK versions that have fewer than four
        //parts, eg, 3.1.2 instead of 3.1.2.0.
        while (result.length() < 8)
        {
            result += "0";
        }
        
        return result;
    }
}
