#
# Mon Jun 26 09:41:21 BST 2017 CafeX Communications
#
// Standard Jenkins environment parameters
def gitUrl = GIT_URL

// Injected parameters from the 'find branches' seed job
branchName = GIT_BRANCH
repository = GIT_REPOSITORY
cafeXBitBucketCredentialsId = CAFEX_BITBUCKET_GIT_CREDENTIALS_ID
folderName = FOLDER_NAME

// TODO: This method should be moved to common Jenkins Job DSL (See Jira GBS-65)
def setBitBucketBuildStatus(String status) {
  // Update the BitBucket build status as the build progresses
  // In the future, this functionality would be better implemented as a Jenkins plugin, for example https://github.com/Flagbit/jenkins-bitbucket-build-status-notifier
  // BitBucket key values for the build can be up to 40 characters so we need to create a hashed value to limit the length
  return 'set +x' + System.lineSeparator() +
          'SHORT_JOB_NAME=`echo ${JOB_NAME} | awk \'{print substr($0,length-29,30)}\'`' + System.lineSeparator() +
          'BITBUCKET_KEY_HASH=`echo ${BUILD_TAG}-${GIT_COMMIT}-${GIT_BRANCH} | sha1sum`' + System.lineSeparator() +
          'BITBUCKET_ID=${BITBUCKET_KEY_HASH:0:9}-${SHORT_JOB_NAME}' + System.lineSeparator() +
          'echo "Setting Bitbucket build status to ' + status + ' for build ID ${BITBUCKET_ID}"' + System.lineSeparator() +
          'curl --silent --show-error -u ' + CAFEX_BITBUCKET_REST_CREDENTIALS + ' -H "Content-Type: application/json" -X POST ' +
          'https://api.bitbucket.org/2.0/repositories/cafexcomms/' + repository + '/commit/${GIT_COMMIT}/statuses/build' +
          ' -d  "{\\"state\\": \\"' + status + '\\",' + '\\"key\\": \\"${BITBUCKET_ID}\\",' +
          '\\"url\\": \\"${BUILD_URL}\\"}" > /dev/null' + '|| true'
}

// Note: Cannot use '/' in job name so use '-' instead
def jobName = folderName + "/" + 'build'

job(jobName) {
  description("Jenkins DSL job for building branch ${branchName} of repository ${repository}")
  label('mac')
  scm {
    git {
      remote {
        name('origin')
        url(gitUrl)
        credentials(cafeXBitBucketCredentialsId)
      }
      branch(branchName)
    }
  }
  triggers {
    scm('H/5 * * * *')
  }
  steps {
    shell(setBitBucketBuildStatus('INPROGRESS'))
    gradle {
      useWrapper(true)
      tasks('clean')
      tasks('buildSocketRocket')
      if (branchName.startsWith('release')) {
        tasks('uploadArchives')
      }
    }
  }
  publishers {
    // Update BitBucket post build status
    postBuildScripts {
      onlyIfBuildFails(true)
      onlyIfBuildSucceeds(false)
      steps {
        shell(setBitBucketBuildStatus('FAILED'))
      }
    }
    postBuildScripts {
      onlyIfBuildSucceeds(true)
      steps {
        shell(setBitBucketBuildStatus('SUCCESSFUL'))
      }
    }

    mailer('mbeale@cafex.com', false, false)
  }
}

println('=== Done ===')
